# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


def make_tab_title_widget(text_cropable, text_fixed='', attr_unfocused='', attr_focused=''):
    import urwid
    from ...utils import strcrop
    from ...tui.main import MAX_TAB_TITLE_WIDTH
    max_width = max(1, MAX_TAB_TITLE_WIDTH-len(text_fixed))
    text_cropped = strcrop(text_cropable, max_width, tail='…')
    return urwid.AttrMap(urwid.Text(''.join((text_cropped, text_fixed))),
                         attr_unfocused, attr_focused)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import urwid
from . import urwidpatches
from ..main import (aioloop, cfg, cmdmgr, srvapi, helpmgr)


#
# Keybindings
#

KEYMAP_CONTEXTS = ('main', 'tabs', 'torrentlist', 'torrent', 'filelist', 'file', 'peerlist')

# Remove urwid's defaults
for key in tuple(urwid.command_map._command):
    del urwid.command_map[key]

from .keymap import Key
urwid.command_map[Key('pgup')]   = urwid.CURSOR_PAGE_UP
urwid.command_map[Key('pgdn')]   = urwid.CURSOR_PAGE_DOWN
urwid.command_map[Key('ctrl-b')] = urwid.CURSOR_PAGE_UP
urwid.command_map[Key('ctrl-f')] = urwid.CURSOR_PAGE_DOWN
urwid.command_map[Key('b')]      = urwid.CURSOR_PAGE_UP
urwid.command_map[Key('space')]  = urwid.CURSOR_PAGE_DOWN

urwid.command_map[Key('up')]     = urwid.CURSOR_UP
urwid.command_map[Key('down')]   = urwid.CURSOR_DOWN
urwid.command_map[Key('left')]   = urwid.CURSOR_LEFT
urwid.command_map[Key('right')]  = urwid.CURSOR_RIGHT
urwid.command_map[Key('meta-b')] = urwid.CURSOR_WORD_LEFT
urwid.command_map[Key('meta-f')] = urwid.CURSOR_WORD_RIGHT

urwid.command_map[Key('home')]   = urwid.CURSOR_MAX_LEFT
urwid.command_map[Key('end')]    = urwid.CURSOR_MAX_RIGHT
urwid.command_map[Key('ctrl-a')] = urwid.CURSOR_MAX_LEFT
urwid.command_map[Key('ctrl-e')] = urwid.CURSOR_MAX_RIGHT

urwid.command_map[Key('ctrl-k')] = urwid.DELETE_TO_EOL
urwid.command_map[Key('ctrl-u')] = urwid.DELETE_LINE
urwid.command_map[Key('ctrl-d')] = urwid.DELETE_CHAR_UNDER_CURSOR
urwid.command_map[Key('meta-d')] = urwid.DELETE_WORD_LEFT
urwid.command_map[Key('meta-backspace')] = urwid.DELETE_WORD_RIGHT
urwid.command_map[Key('ctrl-w')] = urwid.DELETE_WORD_RIGHT

urwid.command_map[Key('enter')]  = urwid.ACTIVATE
urwid.command_map[Key('escape')] = urwid.CANCEL
urwid.command_map[Key('ctrl-g')] = urwid.CANCEL
urwid.command_map[Key('ctrl-l')] = urwid.REDRAW_SCREEN

from ..settings.defaults import DEFAULT_KEYMAP
from .keymap import KeyMap
keymap = KeyMap(callback=lambda cmd,widget: cmdmgr.run_task(cmd, on_error=log.error))
for args in DEFAULT_KEYMAP:
    if args['action'][0] == '<' and args['action'][-1] == '>':
        args['action'] = keymap.mkkey(args['action'])
    keymap.bind(**args)
helpmgr.keymap = keymap


#
# Widgets
#

MAX_TAB_TITLE_WIDTH = 50

from .group import Group
from .tabs import (Tabs, TabBar)
from .cli import CLIEditWidget
from .logger import LogWidget
from .infobar import (KeyChainsWidget, QuickHelpWidget, ConnectionStatusWidget,
                      BandwidthStatusWidget, TorrentCountersWidget)
from . import theme

def load_theme(themeobj):
    """Load theme from `themeobj`

    themeobj: See `theme.load`

    If `themeobj` is a string, does not have a path and does not exist in the
    current working directory, try to load it from the same path as the rc file.
    """
    import os
    if isinstance(themeobj, str) and \
       os.sep not in themeobj and not os.path.exists(themeobj):
        # Path is not given and file does not exist in working dir.
        from ..settings.defaults import DEFAULT_RCFILE
        from ..main import cliargs
        rcfilepath = cliargs['rcfile'] or DEFAULT_RCFILE
        themefilepath = os.path.join(os.path.dirname(rcfilepath), themeobj)
        if os.path.exists(themefilepath):
            theme.load(themefilepath, urwidscreen)
            return
    theme.load(themeobj, urwidscreen)


def _create_cli_widget():
    def on_cancel(widget):
        widget.set_edit_text('')
        widgets.hide('cli')

    def on_accept(widget):
        cmd = widget.get_edit_text()
        on_cancel(widget)
        cmdmgr.run_task(cmd, on_error=log.error)

    return CLIEditWidget(':',
                         on_accept=on_accept, on_cancel=on_cancel,
                         history_file=cfg['tui.cli.history'].value)


def _greedy_spacer():
    return urwid.Padding(urwid.Text(''))


topbar = Group(cls=urwid.Columns)
topbar.add(name='host',   widget=ConnectionStatusWidget(), options='pack')
topbar.add(name='spacer', widget=urwid.AttrMap(_greedy_spacer(), 'topbar'))
topbar.add(name='help',   widget=QuickHelpWidget(), options='pack')

tabs = keymap.wrap(Tabs, context='tabs')(
    tabbar=urwid.AttrMap(TabBar(), 'tabs.unfocused')
)

bottombar = Group(cls=urwid.Columns)
bottombar.add(name='counters', widget=TorrentCountersWidget(), options='pack')
bottombar.add(name='spacer', widget=urwid.AttrMap(_greedy_spacer(), 'bottombar'))
bottombar.add(name='bandwidth', widget=BandwidthStatusWidget(), options='pack')

cli = urwid.AttrMap(_create_cli_widget(), 'cli')

logwidget = LogWidget(maxrows=cfg['tui.log.height'].value,
                      autohide_delay=cfg['tui.log.autohide'].value)

keychains = KeyChainsWidget()
keymap.on_keychain(keychains.update)

widgets = keymap.wrap(Group, context='main')(cls=urwid.Pile)
widgets.add(name='topbar', widget=topbar, options='pack')
widgets.add(name='main', widget=tabs)
widgets.add(name='log', widget=logwidget, options='pack', visible=False)
widgets.add(name='cli', widget=cli, options='pack', visible=False)
widgets.add(name='bottombar', widget=bottombar, options='pack')
widgets.add(name='keychains', widget=keychains, options='pack')


def unhandled_input(key):
    key = keymap.evaluate(key)
    if key is not None:
        log.debug('Unhandled key: %s', key)

urwidscreen = urwid.raw_display.Screen()
urwidloop = urwid.MainLoop(widgets,
                           screen=urwidscreen,
                           event_loop=urwid.AsyncioEventLoop(loop=aioloop),
                           unhandled_input=unhandled_input,
                           handle_mouse=False)


def run(command_runner):
    """Run commands and start TUI

    Return False if any of the commands failed, True otherwise.
    """
    # Don't catch theme.ParserError - a broken default theme should make us
    # croak obviously and horribly
    theme.init(cfg['tui.theme'].default, urwidscreen)

    # Load tui-specific hooks before commands run (commands may trigger hooks)
    from . import hooks

    try:
        if not command_runner():
            return False
    # Make 'quit' behave as expected
    except urwid.ExitMainLoop:
        return True

    # Start logging to TUI widget instead of stdout/stderr
    logwidget.enable()

    topbar.help.update()

    # If no tabs have been opened by cli or rc file, open default tab
    if len(tabs) <= 0:
        for cmd in ( 'tab ls -c size,ratio,seeds,status,tracker,path,name,activity',
                     'tab ls active|incomplete',
                    ('tab ls downloading -c size,downloaded,progress,'
                     'rate-down,completed,eta,path,name'),
                    ('tab ls uploading -c size,uploaded,ratio,'
                     'rate-up,connections,seeds,tracker,path,name -s ratio'),
                    'tab ls stopped -c size,progress,seeds,activity,path,name',
                    'tab ls isolated -c error,tracker,path,name -s tracker',
                    'tab -t peers lsp -s eta,torrent',):
            cmdmgr.run_sync(cmd, on_error=log.error)
        tabs.focus_position = 1

    try:
        # Start polling torrent lists, counters, bandwidth usage, etc.
        aioloop.run_until_complete(srvapi.start_polling())
        urwidloop.run()
    finally:
        logwidget.disable()
        aioloop.run_until_complete(srvapi.stop_polling())

    return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Torrent commands for the TUI"""

from ...logging import make_logger
log = make_logger(__name__)

from collections import abc
from functools import partial

from ..base import torrent as base
from . import mixin
from .. import (ExpectedResource, InitCommand)
from .utils import make_tab_title_widget


class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.polling_frenzy, mixin.make_request):
    provides = {'tui'}


class AnnounceTorrentsCmd(base.AnnounceTorrentsCmdbase,
                          mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.select_torrents, mixin.generate_tab_title):
    provides = {'tui'}
    tui = ExpectedResource

    async def make_tlist(self, tfilter, sort, columns):
        if 'marked' not in columns:
            columns.insert(0, 'marked')

        make_titlew = partial(make_tab_title_widget,
                              attr_unfocused='tabs.torrentlist.unfocused',
                              attr_focused='tabs.torrentlist.focused')

        title_str = await self.generate_tab_title(tfilter)

        from ...tui.torrent.tlist import TorrentListWidget
        tlistw = TorrentListWidget(tfilter=tfilter, sort=sort, columns=columns,
                                   title=title_str)
        tabid = self.tui.tabs.load(make_titlew(tlistw.title), tlistw)

        def set_tab_title(text, count):
            self.tui.tabs.set_title(make_titlew(text, count), position=tabid)
        tlistw.title_updater = set_tab_title

        return True


class ListFilesCmd(base.ListFilesCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.select_files,
                   mixin.generate_tab_title):
    provides = {'tui'}
    tui = ExpectedResource
    srvapi = ExpectedResource

    async def make_flist(self, tfilter, ffilter, columns):
        if 'marked' not in columns:
            columns.insert(0, 'marked')

        make_titlew = partial(make_tab_title_widget,
                              attr_unfocused='tabs.filelist.unfocused',
                              attr_focused='tabs.filelist.focused')

        title_str = await self.generate_tab_title(tfilter)

        from ...tui.torrent.flist import FileListWidget
        flistw = FileListWidget(self.srvapi, tfilter, ffilter, columns,
                                title=title_str)
        tabid = self.tui.tabs.load(make_titlew(flistw.title), flistw)

        def set_tab_title(text, count):
            self.tui.tabs.set_title(make_titlew(text, count), position=tabid)
        flistw.title_updater = set_tab_title

        return True


class ListPeersCmd(base.ListPeersCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.generate_tab_title):
    provides = {'tui'}
    tui = ExpectedResource
    srvapi = ExpectedResource

    async def make_plist(self, tfilter, pfilter, sort, columns):
        make_titlew = partial(make_tab_title_widget,
                              attr_unfocused='tabs.peerlist.unfocused',
                              attr_focused='tabs.peerlist.focused')

        title_str = await self.generate_tab_title(tfilter)

        from ...tui.torrent.plist import PeerListWidget
        plistw = PeerListWidget(self.srvapi, tfilter=tfilter, pfilter=pfilter,
                                sort=sort, columns=columns, title=title_str)
        tabid = self.tui.tabs.load(make_titlew(plistw.title), plistw)

        def set_tab_title(text, count):
            self.tui.tabs.set_title(make_titlew(text, count), position=tabid)
        plistw.title_updater = set_tab_title

        return True


class TorrentDetailsCmd(base.TorrentDetailsCmdbase,
                        mixin.select_torrents, mixin.make_request,
                        mixin.generate_tab_title):
    provides = {'tui'}
    tui = ExpectedResource

    async def show_details(self, tfilter):
        tid = await self.get_torrent_id(tfilter)
        if tid is None:
            return False

        make_titlew = partial(make_tab_title_widget,
                              attr_unfocused='tabs.torrentdetails.unfocused',
                              attr_focused='tabs.torrentdetails.focused')

        title_str = await self.generate_tab_title(tfilter)
        from ...tui.torrent.tdetails import TorrentDetailsWidget
        TorrentDetailsWidget_keymapped = self.tui.keymap.wrap(TorrentDetailsWidget,
                                                              context='torrent')
        detailsw = TorrentDetailsWidget_keymapped(self.srvapi, tid, title=title_str)
        tabid = self.tui.tabs.load(make_titlew(detailsw.title), detailsw)

        def set_tab_title(text):
            self.tui.tabs.set_title(make_titlew(text), position=tabid)
        detailsw.title_updater = set_tab_title

        return True


class SortCmdbase(metaclass=InitCommand):
    name = 'sort'
    aliases = ()
    provides = {'tui'}
    category = 'torrent'
    description = "Sort lists of torrents or peers"
    usage = ('sort [<OPTIONS>] [<ORDER> <ORDER> <ORDER> ...]',)
    examples = ('sort tracker status !rate-down',
                'sort --add eta')
    argspecs = (
        {'names': ('ORDER',), 'nargs': '*',
         'description': 'How to sort list items (see SORT ORDERS section)'},

        {'names': ('--add', '-a'), 'action': 'store_true',
         'description': 'Append ORDERs to current list of sort orders instead of replacing it'},

        {'names': ('--reset', '-r'), 'action': 'store_true',
         'description': 'Go back to sort order that was used when list was created'},

        {'names': ('--none', '-n'), 'action': 'store_true',
         'description': 'Remove all sort orders from the list'},
    )

    def _list_sort_orders(title, sortercls):
        return (title,) + \
            tuple('\t{}\t - \t{}'.format(', '.join((sname,) + s.aliases), s.description)
                  for sname,s in sorted(sortercls.SORTSPECS.items()))

    from ...client.sorters.tsorter import TorrentSorter
    from ...client.sorters.psorter import TorrentPeerSorter
    more_sections = {
        'SORT ORDERS': _list_sort_orders('TORRENT LISTS', TorrentSorter) + \
                       ('',) + \
                       _list_sort_orders('PEER LISTS', TorrentPeerSorter)
    }

    tui = ExpectedResource

    async def run(self, add, reset, none, ORDER):
        current_tab = self.tui.tabs.focus

        if reset:
            current_tab.sort = 'RESET'

        if none:
            current_tab.sort = None

        if ORDER:
            # Find appropriate sorter class for focused list
            from ...tui.torrent.tlist import TorrentListWidget
            from ...tui.torrent.plist import PeerListWidget
            if isinstance(current_tab, TorrentListWidget):
                sortcls = self.TorrentSorter
            elif isinstance(current_tab, PeerListWidget):
                sortcls = self.TorrentPeerSorter
            else:
                log.error('Current tab does not contain a torrent or peer list.')
                return False

            try:
                new_sort = sortcls(ORDER)
            except ValueError as e:
                log.error(e)
                return False

            if add and current_tab.sort is not None:
                current_tab.sort += new_sort
            else:
                current_tab.sort = new_sort
            return True


class MoveTorrentsCmd(base.MoveTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class PriorityCmd(base.PriorityCmdbase,
                  mixin.polling_frenzy, mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}


class RateLimitCmd(base.RateLimitCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.polling_frenzy):
    provides = {'tui'}


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from collections import abc


class TorrentBase(abc.Mapping):
    """Information about a torrent as a mapping

    This is the base class that all API implementations should use.

    Derivatives of this base class must add the methods 'update',
    '__getitem__' and '__iter__'.
    """

    def update(self, raw_torrent):
        raise NotImplementedError()

    def __getitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def __repr__(self):
        r = '<{} #{}'.format(type(self).__name__, self['id'])
        if 'name' in self:
            r += ' ' + repr(self['name'])
        return r + '>'

    def __eq__(self, other):
        if isinstance(other, int):
            return self['id'] == other
        elif isinstance(other, TorrentBase):
            return self['id'] == other['id']
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self['id'] > other['id']

    def __len__(self):
        return len(tuple(iter(self)))

    def __hash__(self):
        return hash(self['id'])


from collections import abc
from . import tkeys
class TorrentFileTreeBase(abc.Mapping):
    """Nested mapping of a Torrent's files"""

    # Distinguish subtrees from files without comparing classes everywhere
    # ("parent" or "leaf")
    nodetype = 'parent'

    def __init__(self, *args, **kwargs):
        raise NotImplementedError()

    @property
    def files(self):
        """Yield all TorrentFiles recursively"""
        for entry in self._items.values():
            if entry.nodetype == 'leaf':
                yield entry
            else:
                yield from entry.files

    @property
    def folders(self):
        """Yield (name, TorrentFileTree) tuples recursively"""
        for name,entry in self._items.items():
            if entry.nodetype == 'parent':
                yield (name, entry)
                yield from entry.folders

    @property
    def path(self):
        return self._path

    def __repr__(self):
        return '<%s %r>' % (type(self).__name__, self._items)

    def __getitem__(self, key):
        return self._items[key]

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import os

from .defaults import DEFAULT_RCFILE


def _tildify(p):
    if p.startswith(os.environ['HOME']):
        return '~' + p[len(os.environ['HOME']):]
    return p


class RcFileError(Exception):
    pass


def read(filepath=DEFAULT_RCFILE):
    """Read list of commands from file"""
    filepath = os.path.expanduser(filepath)
    log.debug('Reading rc file: %r', filepath)
    try:
        with open(filepath, 'r') as f:
            cmdstrs = (line
                       for line in (line.strip() for line in f.readlines())
                       if line and not line.startswith('#'))

    except FileNotFoundError:
        if _tildify(filepath) == _tildify(DEFAULT_RCFILE):
            return ()  # Missing default rc file is not an error
        else:
            raise RcFileError('File not found: {}'.format(_tildify(filepath)))

    except PermissionError as e:
        raise RcFileError('No read permission for rc file: {}'.format(_tildify(filepath)))

    return cmdstrs



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering TorrentFiles by various values"""

from ..tkeys import (TorrentFile, TorrentFilePriority)
from . import (BoolFilterSpec, make_cmp_filter, Filter, FilterChain)

def _make_cmp_filter(*args, **kwargs):
    return make_cmp_filter(TorrentFile.TYPES, *args, **kwargs)


class SingleTorrentFileFilter(Filter):
    DEFAULT_FILTER = 'name'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda f: True,
            aliases=('*',),
            description='All files'),
        'wanted': BoolFilterSpec(
            lambda f: f['is-wanted'],
            description='Wanted files'),
        'complete': BoolFilterSpec(
            lambda f: f['progress'] >= 100,
            description='Fully downloaded files'),
    }

    COMPARATIVE_FILTERS = {
        'name': _make_cmp_filter('name', 'Match VALUE against file name'),
        'path': _make_cmp_filter('path', 'Match VALUE against path in torrent'),
        'size': _make_cmp_filter('size-total', 'Match VALUE against file size'),
        'downloaded': _make_cmp_filter('size-downloaded',
                                       'Match VALUE against number of downloaded bytes',
                                       aliases=('down',)),
        '%downloaded': _make_cmp_filter('progress',
                                        'Match VALUE against percentage of downloaded bytes',
                                        aliases=('%down',)),
        'priority': _make_cmp_filter('priority',
                                     'Match VALUE against download priority (low, normal, high or shun)',
                                     aliases=('prio',)),
    }


class TorrentFileFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTorrentFileFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from .settings import (ValueBase, StringValue, IntegerValue, NumberValue,
                       BooleanValue, PathValue, ListValue, SetValue,
                       OptionValue)
from ..client.errors import ClientError
from ..client import constants as const
from ..client import convert


class SrvValueBase(ValueBase):
    def __init__(self, name, getter, setter, description='No description available'):
        super().__init__(name, default=const.DISCONNECTED, description=description)
        self._getter = getter
        self._setter = setter

    @property
    def value(self):
        return self._getter()

    async def set(self, value):
        log.debug('Setting server value %r: %r', self.name, value)
        try:
            value = self.convert(value)
            self.validate(value)
            await self._setter(value)
        except ClientError as e:
            raise ValueError("Can't change server setting {}: {}".format(self.name, e))
        except ValueError as e:
            log.debug('%s while setting %s to %r: %s', type(e).__name__, self.name, value, e)
            raise ValueError('{} = {}: {}'.format(self.name, self.str(value), e))
        else:
            log.debug('Successfully set %s to %r', self.name, value)

    def convert(self, value):
        log.debug('SrvValueBase: converting %r', value)
        if value is const.DISCONNECTED:
            return value
        else:
            log.debug('consulting super().convert for %r', value)
            value = super().convert(value)
            log.debug('got back: %r', value)
            return value

    def validate(self, value):
        log.debug('SrvValueBase: validating %r', value)
        if value is not const.DISCONNECTED:
            log.debug('consulting super().validate for %r', value)
            super().validate(value)
            log.debug('Valid value: %r', value)


class RateLimitSrvValue(SrvValueBase, NumberValue):
    typename = 'number or bool'
    valuesyntax = ('[+=|-=]<NUMBER>[k|M|G|T|Ki|Mi|Gi|Ti][b|B] or '
                   '%s (case is ignored)' % BooleanValue.valuesyntax)

    def convert(self, value):
        def convert_bandwidth(value):
            try:
                return convert.bandwidth.from_string(value)
            except ValueError as e:
                raise ValueError('Not a {}: {!r}'.format(self.typename, value))

        try:
            # value may be something like 'on' or 'off'
            return BooleanValue.convert(self, value)
        except ValueError:
            # Parse relative values
            if isinstance(value, str) and len(value) >= 3 and value[:2] in ('+=', '-='):
                op = value[:2]
                num = convert_bandwidth(value[2:].strip())

                # Pretend current value is 0 if user wants to adjust UNLIMITED.
                if self.value is const.UNLIMITED and op == '+=':
                    return num

                return super().convert(op + str(float(num)))

            # Parse other strings or numbers
            elif not const.is_constant(value):
                return convert_bandwidth(value)

            # Let parent provide the error message
            return super().convert(value)

    def validate(self, value):
        if isinstance(value, bool) or value in (const.ENABLED, const.DISABLED):
            pass
        else:
            super().validate(value)  # May raise ValueError


class PathSrvValue(SrvValueBase, PathValue):
    pass

class PathIncompleteSrvValue(PathSrvValue):
    typename = 'path or bool'
    def convert(self, value):
        try:
            # value may be something like 'on' or 'off'
            return BooleanValue.convert(self, value)
        except ValueError:
            return super().convert(value)


def is_server_setting(name):
    """Whether setting `name` is managed by the server"""
    if isinstance(name, ValueBase):
        name = name.name
    else:
        name = str(name)

    if name.startswith('srv.') and \
       not name.startswith('srv.timeout') and \
       not name.startswith('srv.url'):
        return True
    else:
        return False



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Commands that work exclusively in the TUI"""

from ...logging import make_logger
log = make_logger(__name__)


from . import make_tab_title_widget
from .. import (InitCommand, ExpectedResource)


from ...tui.main import KEYMAP_CONTEXTS
import shlex
class BindCmd(metaclass=InitCommand):
    name = 'bind'
    provides = {'tui'}
    category = 'tui'
    description = 'Bind keys to commands or other keys'
    usage = ('bind [<OPTIONS>] <KEY> <ACTION>',)
    examples = ('bind --context tabs alt-[ tab --focus left',
                'bind --context tabs alt-] tab --focus right',
                'bind --context torrent alt-! start --force',
                'bind ctrl-a tab ls active',
                "bind 'd .' delete",
                "bind 'd !' delete --delete-files",
                'bind u <up>',
                'bind d <down>')
    argspecs = (
        { 'names': ('--context','-c'),
          'description': 'Where KEY is grabbed (see CONTEXTS section)' },
        { 'names': ('KEY',),
          'description': 'One or more keys or key combinations (see KEYS section)' },
        { 'names': ('ACTION',), 'nargs': 'REMAINDER',
          'description': ("Any command or '<KEY>' (including the brackets) "
                          'to translate one key to another') },
    )

    def __create_CONTEXTS_section():
        lines = [
            ('Keys are mapped in contexts.  With no context given, the default '
             'context is used.  The default context only gets the key if no '
             'other context wants it.  The same key can be mapped to different '
             'actions in different contexts.'),
            '',
            'Available contexts are: ' + \
              ', '.join('%s' % context for context in KEYMAP_CONTEXTS),
            '',
            'EXAMPLE',
            '\tbind --context torrent ctrl-t start',
            '\tbind --context tabs ctrl-t tab',
            '\tbind ctrl-t <left>',
            '',
            ('\tWhen focusing a torrent, <ctrl-t> starts the focused torrent.  '
             'If focus is not on a torrent but still on a tab (e.g. when reading '
             'documentation) a new tab is opened.  Otherwise (e.g. focus is on the '
             'command prompt), <ctrl-t> does the same as <left>.'),
        ]
        return lines

    more_sections = {
        'CONTEXTS': __create_CONTEXTS_section,
        'KEYS': (
            'Single-character keys are specified as themselves (e.g. h, X, 5, !, þ, ¥, etc).',
            '',
            ('Special key names are enter, space, tab, backspace, insert, delete, home, end, '
             'up, down, left, right, pgup, pgdn and f1-12.'),
            '',
            ("The modifiers 'ctrl', 'alt' and 'shift' are separated with '-' from the key "
             "(e.g. alt-i, shift-delete, ctrl-a).  shift-x is identical to X."),
            '',
            ("Chained keys are sparated by single spaces (' ') or pluses ('+') and must be "
             "given as a single argument."),
        )
    }

    tui = ExpectedResource

    def run(self, context, KEY, ACTION):
        keymap = self.tui.keymap

        key = KEY
        if len(ACTION) == 1 and ACTION[0][0] == '<' and ACTION[0][-1] == '>':
            # ACTION is another key (e.g. 'j' -> 'down')
            action = keymap.mkkey(ACTION[0])
        else:
            action = ' '.join(shlex.quote(x) for x in ACTION)

        if context is not None and context not in KEYMAP_CONTEXTS:
            log.error('Invalid context: {!r}'.format(context))
            return False
        try:
            keymap.bind(key, action, context=context)
        except ValueError as e:
            log.error(e)
            return False
        else:
            return True


class UnbindCmd(metaclass=InitCommand):
    name = 'unbind'
    provides = {'tui'}
    category = 'tui'
    description = 'Unbind keys so pressing them has no effect'
    usage = ('unbind [<OPTIONS>] <KEY> <KEY> <KEY> ...',)
    examples = ('unbind --context main ctrl-l',
                'unbind q')
    argspecs = (
        { 'names': ('--context','-c'),
          'description': 'Where KEY is grabbed (see "bind" command)' },
        { 'names': ('--all','-a'), 'action': 'store_true',
          'description': 'Remove all existing keybindings, including defaults' },
        { 'names': ('KEY',), 'nargs': 'REMAINDER',
          'description': 'Keys or key combinations (see "bind" command)' },
    )

    tui = ExpectedResource

    def run(self, context, all, KEY):
        keymap = self.tui.keymap

        if all:
            keymap.clear()

        if context is not None and context not in KEYMAP_CONTEXTS:
            log.error('Invalid context: {!r}'.format(context))
            return False

        success = True
        for key in KEY:
            try:
                keymap.unbind(key, context=context)
            except ValueError as e:
                log.error(e)
                success = False
            else:
                success = success and True
        return success


class MarkCmd(metaclass=InitCommand):
    name = 'mark'
    provides = {'tui'}
    category = 'tui'
    description = 'Select torrents or files for an action'
    usage = ('mark [<OPTIONS>]',)
    argspecs = (
        { 'names': ('--focus-next','-n'), 'action': 'store_true',
          'description': 'Move focus forward after marking or toggling' },
        { 'names': ('--toggle','-t'), 'action': 'store_true',
          'description': 'Mark if unmarked, unmark if marked' },
        { 'names': ('--all','-a'), 'action': 'store_true',
          'description': 'Mark or toggle all items' },
    )
    more_sections = {
        'NOTES': (('The column "marked" must be in the "columns.*" settings. Otherwise '
                   'marked list items are indistinguishable from unmarked ones.'),
                  '',
                  ('The character that is displayed in the "marked" column is '
                   'specified by the settings "tui.marked.on" and "tui.marked.off".')),
    }

    tui = ExpectedResource

    def run(self, focus_next, toggle, all):
        widget = self.tui.tabs.focus
        if not hasattr(widget, 'mark'):
            log.error('Nothing to mark here.')
            return False
        else:
            widget.mark(toggle=toggle, all=all)
            if focus_next:
                widget.focus_position += 1
            return True


class UnmarkCmd(metaclass=InitCommand):
    name = 'unmark'
    provides = {'tui'}
    category = 'tui'
    description = 'Deselect torrents or files for an action'
    usage = ('unmark [<OPTIONS>]',)
    argspecs = (
        { 'names': ('--focus-next','-n'), 'action': 'store_true',
          'description': 'Move focus forward after unmarking or toggling' },
        { 'names': ('--toggle','-t'), 'action': 'store_true',
          'description': 'Mark if unmarked, unmark if marked' },
        { 'names': ('--all','-a'), 'action': 'store_true',
          'description': 'Unmark or toggle all items' },
    )
    more_sections = MarkCmd.more_sections

    tui = ExpectedResource

    def run(self, focus_next, toggle, all):
        widget = self.tui.tabs.focus
        if not hasattr(widget, 'unmark'):
            log.error('Nothing to unmark here.')
            return False
        else:
            widget.unmark(toggle=toggle, all=all)
            if focus_next:
                widget.focus_position += 1
            return True


class ClearLogCmd(metaclass=InitCommand):
    name = 'clearlog'
    provides = {'tui'}
    category = 'tui'
    description = 'Clear all logged messages'

    tui = ExpectedResource

    def run(self):
        logwidget = self.tui.logwidget
        if len(tuple(logwidget.entries)) > 0:
            logwidget.clear()
            return True
        else:
            return False


class QuitCmd(metaclass=InitCommand):
    name = 'quit'
    provides = {'tui'}
    category = 'tui'
    description = 'Terminate the TUI'

    def run(self):
        import urwid
        raise urwid.ExitMainLoop()


class RcCmd(metaclass=InitCommand):
    name = 'rc'
    aliases = ('source',)
    category = 'misc'
    provides = {'tui'}
    description = 'Run commands in rc file'
    usage = ('rc <FILE>',)
    examples = ('rc rc.example   # Load $XDG_CONFIG_HOME/.config/{APPNAME}/rc.example',)
    argspecs = (
        {'names': ('FILE',),
         'description': ('Path to rc file; if FILE does not start with '
                         "'.', '~' or '/', $XDG_CONFIG_HOME/.config/{APPNAME}/ "
                         'is prepended')},
    )
    cmdmgr = ExpectedResource

    async def run(self, FILE):
        from ...settings import rcfile
        from ...settings.defaults import DEFAULT_RCFILE
        import os

        if FILE[0] not in (os.sep, '.', '~'):
            FILE = '{}{}{}'.format(os.path.dirname(DEFAULT_RCFILE),
                                   os.sep,
                                   FILE)

        try:
            lines = rcfile.read(FILE)
        except rcfile.RcFileError as e:
            log.error('Loading rc file failed: {}'.format(e))
            return False
        else:
            log.debug('Running commands from rc file: %r', FILE)
            for cmdline in lines:
                log.debug('  %r', cmdline)
                success = await self.cmdmgr.run_async(cmdline)
                if not success:
                    return False
            return True


class TabCmd(metaclass=InitCommand):
    name = 'tab'
    provides = {'tui'}
    category = 'tui'
    description = 'Open, close and focus tabs'
    usage = ('tab [<OPTIONS>] [<COMMAND>]',)
    examples = ('tab',
                'tab ls active',
                'tab -b ls active',
                'tab -f active',
                'tab -f 3 ls active',
                'tab -c')
    argspecs = (
        { 'names': ('--background', '-b'), 'default': False, 'action': 'store_true',
          'description': 'Opens tab in background, instead of focusing it' },
        { 'names': ('--close', '-c'), 'nargs': '?', 'const': None, 'default': False,
          'description': 'Close tab at index or with partial title CLOSE' },
        { 'names': ('--close-all', '-ca'), 'default': False, 'action': 'store_true',
          'description': 'Close all tabs' },
        { 'names': ('--focus', '-f'),
          'description': ('Focus tab; FOCUS can be an index (first tab is 1), '
                          'part of a tab title, "left" or "right"') },
        { 'names': ('--title', '-t'),
          'description': 'Manually set tab title instead of generating one' },
        { 'names': ('COMMAND',), 'nargs': 'REMAINDER',
          'description': ('Command to run in new tab') },
    )

    tui = ExpectedResource
    cmdmgr = ExpectedResource

    async def run(self, close, close_all, focus, background, title, COMMAND):
        # Get indexes before adding/removing tabs, which changes indexes on
        # subsequent operations
        if focus is not None:
            i_focus = self.get_tab_index(focus)
            log.debug('Focusing tab %r at index %r', focus, i_focus)
            if i_focus is None:
                return False
        if close is not False:
            i_close = self.get_tab_index(close)
            log.debug('Closing tab %r at index %r', close, i_close)
            if i_close is None:
                return False

        tabs = self.tui.tabs
        old_index = tabs.focus_position

        # Apply close/focus operations
        if focus is not None:
            tabs.focus_position = i_focus
        if close_all is not False:
            tabs.clear()
        elif close is not False:
            tabs.remove(i_close)

        # Remember which torrent is focused in current tab so we can provide
        # it to the command running in a new tab.
        cmdargs = {'title': title} if title else {}
        try:
            cmdargs['focused_torrent_id'] = tabs.focus.focused_torrent_id
        except AttributeError:
            pass

        if close is False and close_all is False and focus is None:
            titlew = make_tab_title_widget(title or 'Empty tab',
                                           attr_unfocused='tabs.unfocused',
                                           attr_focused='tabs.focused')
            tabs.insert(titlew, position='right')
            log.debug('Inserted new tab at position %d: %r', tabs.focus_position, titlew.base_widget.text)

        if COMMAND:
            cmd = ' '.join(shlex.quote(arg) for arg in COMMAND)
            log.debug('Running command in tab %d with args %s: %r',
                      tabs.focus_position,
                      ', '.join('%s=%r' % (k,v) for k,v in cmdargs.items()),
                      cmd)

            cmd = await self.cmdmgr.run_async(cmd, **cmdargs)
            retval = cmd
        else:
            retval = True

        if background:
            tabs.focus_position = old_index

        return retval

    def get_tab_index(self, pos):
        tabs = self.tui.tabs

        if pos is None:
            return tabs.focus_position

        def find_index(pos):
            i = pos-1 if pos > 0 else pos
            try:
                return tabs.get_index(i)
            except IndexError as e:
                log.error('No tab at position: {}'.format(pos))

        def find_title(string):
            for i,title in enumerate(tabs.titles):
                if string in title.original_widget.text:
                    return i
            log.error('No tab found: {!r}'.format(string))

        tabcount = len(tabs)
        curpos = tabs.focus_position
        curpos = 1 if curpos is None else curpos

        try:
            return find_index(int(pos))
        except ValueError:
            if pos == 'left':
                if tabcount > 1:
                    return tabs.get_index(max(0, curpos-1))
            elif pos == 'right':
                if tabcount > 1:
                    return tabs.get_index(min(tabcount-1, curpos+1))
            else:
                return find_title(pos)


class TUICmd(metaclass=InitCommand):
    name = 'tui'
    provides = {'tui'}
    category = 'tui'
    description = 'Show or hide parts of the text user interface'
    usage = ('tui <ACTION> <ELEMENT> <ELEMENT> ...',)
    examples = ('tui toggle log',
                'tui hide topbar.help')
    argspecs = (
        { 'names': ('ACTION',), 'choices': ('show', 'hide', 'toggle'),
          'description': '"show", "hide" or "toggle"' },
        { 'names': ('ELEMENT',), 'nargs': '+',
          'description': ('Name of TUI elements; '
                          'see ELEMENT NAMES section for a list') },
    )

    # Lazily load the element names from tui (HelpManager supports sequences
    # of lines or a callable that returns them)
    def __create_element_names():
        from ...tui.main import widgets
        return ('Available TUI element names are: ' + \
                ', '.join(str(e) for e in sorted(widgets.names_recursive)),)
    more_sections = { 'ELEMENT NAMES': __create_element_names }

    tui = ExpectedResource

    def run(self, ACTION, ELEMENT):
        widgets = self.tui.widgets
        widget = None
        success = False
        for element in ELEMENT:
            # Resolve path
            path = element.split('.')
            target_name = path.pop(-1)
            current_path = []
            widget = widgets
            try:
                for widgetname in path:
                    current_path.append(widgetname)
                    widget = getattr(widget, widgetname)
            except AttributeError as e:
                log.error('Unknown TUI element: %r', '.'.join(current_path))

            if widget is not None:
                action = getattr(widget, ACTION)
                log.debug('%sing %s in %s', ACTION.capitalize(), target_name, widget)
                try:
                    action(target_name)
                    success = True
                except ValueError as e:
                    log.error(e)
        return success



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid
import operator


class _FlowFill(urwid.SolidFill):
    def rows(self, size, focus=False):
        return 0


class Group(urwid.WidgetWrap):
    """Wrapper aroung Pile or Columns widget

    The purpose of this class it o make adding/removing, hiding/showing and
    accessing widgets simpler.
    """

    def __init__(self, *widgets, cls=urwid.Columns, **kwargs):
        """Create new Group widget

        Widgets can be added by providing mappings as positional arguments.
        Each mapping is then provided to the `add` method as keyword
        arguments.

        cls: `Columns` or `Pile` (or derivatives of either)

        All other keyword arguments are forwarded to `cls` on instantiation.
        """
        self._main = cls([], **kwargs)
        self._items = []
        # Add initial widgets
        for widget in widgets:
            self.add(**widget)
        super().__init__(self._main)

    def _get_item(self, name=None, position=None, visible=False):
        """Return item dict identified by `name` or `position`

        visible: If True, return None instead of widget if widget is hidden

        Raises ValueError in case of invalid name or IndexError in case of
        invalid position.
        """
        if name is not None:
            for item in self._items:
                if item['name'] == name:
                    if not visible or self.visible(item['name']):
                        return item
                    else:
                        return None
            raise ValueError('Unknown name: {}'.format(name))

        elif position is not None:
            try:
                item = self._items[position]
            except IndexError:
                raise IndexError('No item at position: {}'.format(position))
            else:
                if not visible or self.visible(item['name']):
                    return item
                else:
                    return None
        else:
            raise TypeError('Neither name nor position specified')

    def _get_position(self, name, visible=False):
        """Return position of item

        visible: If True, return None instead of widget if widget is hidden

        Raises ValueError if no widget with `name` exists.
        """
        try:
            item = self._get_item(name)
        except ValueError:
            raise ValueError('Unknown name: {}'.format(name))
        else:
            if visible:
                content = (item['widget'], item['options'])
                try:
                    return self._main.contents.index(content)
                except ValueError:
                    return None
            else:
                return self._items.index(item)

    def _parse_options(self, opts):
        """Convert sizing options from a simpler format to urwid format

        See set_size method.
        """
        if type(opts) is tuple:
            # Assume tuple is following urwid's format
            return self._main.options(*opts)
        elif type(opts) is str and opts.isdigit():
            return self._main.options('weight', int(opts))
        elif opts == 'pack':
            return self._main.options('pack', None)
        elif type(opts) is int:
            return self._main.options('given', opts)
        else:
            raise ValueError('Invalid options: {}'.format(opts))

    def add(self, name, widget, options=('weight', 100),
            position='end', visible=True, removable=False):
        """Insert new widget

        name: A string ([a-zA-Z0-9_]+) to get the widget via an attribute

        widget: Any widget that can live in a Columns/Pile
        options: See `set_size` method
        position: Insert position (integer, 'start' or 'end')
        visible: True to immediately show widget, False otherwise
        removable: True to allow complete removal of widget, False otherwise

        Raises ValueError if `name` already exists.
        """
        options = self._parse_options(options)
        item = dict(
            name = name,           # String handle
            widget = widget,       # Bare widget
            options = options,     # urwid options tuple, e.g. ('given',10) or ('weight',50)
            removable = removable, # Wether this item can be deleted
        )

        if self.exists(name):
            raise ValueError('Already added: {!r}'.format(name))
        else:
            if position == 'start':
                position = 0
            elif position == 'end':
                position = len(self._items)

            self._items.insert(position, item)
            self.show(name)
            if not visible:
                self.hide(name)

    def remove(self, name):
        """Remove widget from group"""
        if self.exists(name):
            position = self._get_position(name)
            item = self._get_item(name)
            if not item['removable']:
                raise ValueError('Item is not removable: {}'.format(name))
            self._main.contents.pop(position)
            self._items.remove(item)
        else:
            raise ValueError('Unknown item name: {}'.format(name))

    def clear(self):
        """Remove all removable items"""
        for item in tuple(self._items):
            if item['removable']:
                self.remove(item['name'])

    def replace(self, name, widget):
        """Remove `name` if it exists and add new item with the same name"""
        if not self.exists(name):
            raise ValueError('Unknown item name: {}'.format(name))
        else:
            visible = self.visible(name)
            item = self._get_item(name=name)
            item['widget'] = widget
            self.hide(name)
            if visible:
                self.show(name)

    def set_size(self, name, opts):
        """Change size options for widget

        opts: Must be one of the following:
                - An integer is translated to ('given', int).
                - A string that consists solely of numbers is translated to
                  ('weight', int).
                - The string 'pack' is translated to ('pack', None).
                - Any tuple Pile/Columns accepts as 'options' when adding to
                  the contents attribute.
        """
        item = self._get_item(name=name)
        item['options'] = self._parse_options(opts)
        self.hide(name)  # Refresh content in self._main
        self.show(name)

    def show(self, name):
        """Show widget with specific name and focus it if selectable"""
        if not self.exists(name):
            raise ValueError('Unknown item name: {}'.format(name))
        elif not self.visible(name):
            item = self._get_item(name)
            position = self._get_position(name)
            content = (item['widget'], item['options'])
            if position >= len(self._main.contents):
                self._main.contents.append(content)
            else:
                self._main.contents[position] = content

            if item['widget'].selectable():
                self.focus_name = item['name']

    def hide(self, name):
        """Hide widget with specific name and focus the next selectable widget"""
        if not self.exists(name):
            raise ValueError('Unknown item name: {!r}'.format(name))
        elif self.visible(name):
            position = self._get_position(name)
            opts = self._get_item(name)['options']

            if opts[0] == 'weight':
                self._main.contents[position] = (_FlowFill(),
                                                 self._main.options('weight', opts[1]))
            else:
                self._main.contents[position] = (_FlowFill(),
                                                 self._main.options('given', 0))
            # Try to focus next selectable item
            self.focus_selectable(forward=False)
            self.focus_selectable(forward=True)

    def toggle(self, name):
        """Show widget if it's hidden and vice versa"""
        if self.exists(name):
            self.hide(name) if self.visible(name) else self.show(name)
        else:
            raise ValueError('Unknown item name: {!r}'.format(name))

    def visible(self, name):
        """Whether widget is hidden or not"""
        item = self._get_item(name)
        content = (item['widget'], item['options'])
        return content in self._main.contents

    def exists(self, name):
        """Whether widget exists or not"""
        return any(item['name'] == name for item in self._items)

    def __getattr__(self, name):
        """Return widget"""
        try:
            return self._get_item(name)['widget']
        except ValueError as e:
            raise AttributeError(e)

    @property
    def names(self):
        """Return list of known widget names"""
        return [item['name'] for item in self._items]

    @property
    def names_recursive(self):
        """Return list of known widget names recursively

        This dives into Group instances and adds their names, prepending the
        parent's name with '.' as a separator.
        """
        names = []
        for item in self._items:
            if isinstance(item['widget'], Group):
                names.append(item['name'])
                names.extend('{}.{}'.format(item['name'], subname)
                             for subname in item['widget'].names_recursive)
            else:
                names.append(item['name'])
        return names

    @property
    def widgets(self):
        """Return list of all widgets (hidden or visible)"""
        return [item['widget'] for item in self._items]

    @property
    def focus(self):
        """Focused widget or None"""
        try:
            item = self._get_item(position=self._main.focus_position)
            return item['widget']
        except IndexError:
            return None

    @property
    def focus_position(self):
        """Position of currently focused widget or None"""
        return self._main.focus_position

    @focus_position.setter
    def focus_position(self, position):
        self._main.focus_position = position

    @property
    def focus_name(self):
        """Name of currently focused widget or None"""
        try:
            item = self._get_item(position=self._main.focus_position)
            return item['name']
        except IndexError as e:
            return None

    @focus_name.setter
    def focus_name(self, name):
        if self.exists(name):
            position = self._get_position(name, visible=True)
            if position is not None:
                self._main.focus_position = position
        else:
            raise ValueError('Unknown item name: {}'.format(name))

    def focus_selectable(self, forward=True):
        """Change focus to next selectable widget

        forward: True to select next widget, False to select previous widget

        Returns True if focus was changed, False otherwise.
        """
        op = operator.add if forward else operator.sub

        new_pos = op(self.focus_position, 1)
        try:
            while True:
                item = self._get_item(position=new_pos, visible=True)
                if item is not None and item['widget'].selectable():
                    break
                else:
                    new_pos = op(new_pos, 1)
            self.focus_position = new_pos
        except IndexError:
            return False
        else:
            return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for torrent details sections"""

from ..logging import make_logger
log = make_logger(__name__)

from functools import partial


def _size_hr(t):
    if t['size-total'] == t['size-final']:
        return '%s' % t['size-total'].with_unit
    else:
        return '%s (%s wanted)' % (t['size-total'].with_unit, t['size-final'].with_unit)

def _size_mr(t):
    return '%d\t%d' % (t['size-total'], t['size-final'])


def _file_counts(t):
    files = tuple(t['files'].files)
    all_files = len(files)
    wanted_files = sum(1 for f in files if f['is-wanted'])
    return (all_files, wanted_files)

def _files_hr(t):
    return '%d (%d wanted)' % _file_counts(t)

def _files_mr(t):
    return '%d\t%d' % _file_counts(t)


def _pieces_hr(t):
    return '%d * %s' % (t['count-pieces'], t['size-piece'].with_unit)

def _pieces_mr(t):
    return '%d\t%d' % (t['count-pieces'], t['size-piece'])


def _private_hr(t):
    return ('yes (decentralized peer discovery is disabled for this torrent)'
            if t['private'] else
            'no (decentralized peer discovery allowed if enabled globally)')

def _private_mr(t):
    return 'yes' if t['private'] else 'no'


def _uploaded_hr(t):
    return '%s (%.2f %%)' % (t['size-uploaded'].with_unit, t['%uploaded'])

def _uploaded_mr(t):
    return '%d\t%f' % (t['size-uploaded'], t['%uploaded'] / 100)


def _downloaded_hr(t):
    info = ['%.2f %%' % t['%downloaded']]
    if t['%downloaded'] < 100:
        if t['timespan-eta']:
            info.append('%s / %s left' % (t['size-left'].with_unit, t['timespan-eta']))
        else:
            info.append('%s left' % t['size-left'].with_unit)
    return '%s (%s)' % (t['size-downloaded'].with_unit, ', '.join(info))

def _downloaded_mr(t):
    return '%d\t%f\t%d\t%d' % (t['size-downloaded'],
                               t['%downloaded'] / 100, t['size-left'],
                               t['timespan-eta'])


def _ratio_hr(t):
    return '%.4f' % t['ratio']
_ratio_mr = _ratio_hr


def _available_hr(t):
    return '%s (%.2f %%)' % (t['size-available'].with_unit, t['%available'])

def _available_mr(t):
    return '%d\t%f' % (t['size-available'], t['%available'])


def _isolated_hr(t):
    status = t['status']
    return ('yes (torrent can\'t discover new peers)'
            if status.ISOLATED in status else
            'no (torrent can discover new peers)')

def _isolated_mr(t):
    status = t['status']
    return 'yes' if status.ISOLATED in status else 'no'


def _date_hr(key, t):
    date = t[key]
    if date.is_known:
        delta = date.delta
        if abs(delta) < 5:
            return 'now'
        else:
            return '%s (%s)' % (date.full, delta.with_preposition)
    else:
        return date.full

def _date_mr(key, t):
    date = t[key]
    return '%d\t%d' % (date, date.delta)


class Item():
    def __init__(self, label, needed_keys, human_readable=None, machine_readable=None):
        self.label = label
        self.needed_keys = needed_keys
        if human_readable is None:
            self.human_readable = lambda torrent, key=needed_keys[0]: str(torrent[key])
        else:
            self.human_readable = human_readable
        if machine_readable is None:
            self.machine_readable = self.human_readable
        else:
            self.machine_readable = machine_readable


SECTIONS = (
    {'title': 'Torrent', 'width': 60, 'items': (
        Item('Name',       ('name',)),
        Item('ID',         ('id',)),
        Item('Hash',       ('hash',)),
        Item('Size',       ('size-total', 'size-final'), _size_hr, _size_mr),
        Item('Files',      ('files',), _files_hr, _files_mr),
        Item('Pieces',     ('count-pieces', 'size-piece'), _pieces_hr, _pieces_mr),
        Item('Private',    ('private',), _private_hr, _private_mr),
        Item('Comment',    ('comment',)),
        Item('Creator',    ('creator',)),
    )},

    {'title': 'Status', 'width': 51, 'items': (
        Item('Location',   ('path',),),
        Item('Available',  ('%available', 'size-available'), _available_hr, _available_mr),
        Item('Downloaded', ('size-downloaded', 'size-left', '%downloaded', 'timespan-eta'), _downloaded_hr, _downloaded_mr),
        Item('Uploaded',   ('size-uploaded', 'size-total', '%uploaded'), _uploaded_hr, _uploaded_mr),
        Item('Ratio',      ('ratio',), _ratio_hr, _ratio_mr),
        Item('Isolated',   ('status',), _isolated_hr, _isolated_mr),
        Item('Error',      ('error',)),
    )},

    {'title': 'Peers', 'width': 18, 'items': (
        Item('Seeding',     ('peers-seeding',)),
        Item('Connected',   ('peers-connected',)),
        Item('Uploading',   ('peers-uploading',)),
        Item('Downloading', ('peers-downloading',)),
    )},

    {'title': 'Dates', 'width': 39, 'items': (
        Item('Created',    ('time-created',),   partial(_date_hr, 'time-created'),   partial(_date_mr, 'time-created')),
        Item('Added',      ('time-added',),     partial(_date_hr, 'time-added'),     partial(_date_mr, 'time-added')),
        Item('Started',    ('time-started',),   partial(_date_hr, 'time-started'),   partial(_date_mr, 'time-started')),
        Item('Completed',  ('time-completed',), partial(_date_hr, 'time-completed'), partial(_date_mr, 'time-completed')),
        Item('Active',     ('time-activity',),  partial(_date_hr, 'time-activity'),  partial(_date_mr, 'time-activity')),
    )},
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Base classes for configuration commands"""

from ...logging import make_logger
log = make_logger(__name__)

from .. import (InitCommand, ExpectedResource, utils)
from asyncio import iscoroutinefunction


class ResetCmdbase(metaclass=InitCommand):
    name = 'reset'
    category = 'misc'
    provides = set()
    description = 'Reset settings to their default values'
    usage = ('reset <NAME> <NAME> <NAME> ...',)
    examples = ('reset srv.url',)
    argspecs = (
        {'names': ('NAME',), 'nargs': '+',
         'description': 'Name of setting'},
    )
    cfg = ExpectedResource

    def run(self, NAME):
        from ...settings import is_server_setting
        success = True
        for name in NAME:
            if name not in self.cfg:
                log.error('Unknown setting: {}'.format(name))
                success = False
            elif is_server_setting(name):
                log.error('Server settings cannot be reset: {}'.format(name))
                success = False
            else:
                self.cfg[name] = None
        return success


class SetCmdbase(metaclass=InitCommand):
    name = 'set'
    category = 'misc'
    provides = set()
    description = 'Change values of settings'
    usage = ('set <NAME> <VALUE>',)
    examples = ('set srv.url my.server.example.org:12345',)
    argspecs = (
        {'names': ('NAME',),
         'description': 'Name of setting'},
        {'names': ('VALUE',), 'nargs': 'REMAINDER',
         'description': 'New value'},
    )
    cfg = ExpectedResource
    srvapi = ExpectedResource

    async def run(self, NAME, VALUE):
        if NAME not in self.cfg:
            log.error('Unknown setting: {}'.format(NAME))
        else:
            setting = self.cfg[NAME]

            # Make sure strings are strings and lists are lists
            if setting.typename in ('list', 'set'):
                val = utils.listify_args(VALUE)
            else:
                val = ' '.join(VALUE)

            from ...settings import is_server_setting
            if is_server_setting(setting):
                from ...client import ClientError
                try:
                    await self.srvapi.settings.update()
                except ClientError as e:
                    log.error(str(e))
                    return False

            try:
                # Server settings' set() methods are async (maybe others too)
                if iscoroutinefunction(setting.set):
                    await setting.set(val)
                else:
                    setting.set(val)
            except ValueError as e:
                log.error(e)
            else:
                return True

        return False



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column classes that display information in torrent lists

Adding a new cell type adds a new optional column to torrent lists.  To do
this, write a new class that derives from CellWidgetBase and register it in
the COLUMNS dictionary to make it available.
"""

from ...logging import make_logger
log = make_logger(__name__)

import urwid

from ..table import ColumnHeaderWidget
from . import (Style, CellWidgetBase)
from ...views.tlist import COLUMNS as _COLUMNS
from ...client import tkeys


TUICOLUMNS = {}


from .common_columns import MarkedBase
class Marked(MarkedBase):
    style = Style(prefix='torrentlist.marked', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(), style.attrs('header'))

TUICOLUMNS['marked'] = Marked


class Path(_COLUMNS['path'], CellWidgetBase):
    width = 20
    style = Style(prefix='torrentlist.path', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['path'].header),
                           style.attrs('header'))

TUICOLUMNS['path'] = Path


class Connections(_COLUMNS['connections'], CellWidgetBase):
    style = Style(prefix='torrentlist.connections', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['connections'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else ''

TUICOLUMNS['connections'] = Connections


class Seeds(_COLUMNS['seeds'], CellWidgetBase):
    style = Style(prefix='torrentlist.seeds', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['seeds'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 5 else ''

TUICOLUMNS['seeds'] = Seeds


class Progress(_COLUMNS['progress'], CellWidgetBase):
    style = Style(prefix='torrentlist.progress', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['progress'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else ''

TUICOLUMNS['progress'] = Progress


class PercentAvailable(_COLUMNS['%available'], CellWidgetBase):
    style = Style(prefix='torrentlist.%available', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['%available'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else ''

TUICOLUMNS['%available'] = PercentAvailable


class Ratio(_COLUMNS['ratio'], CellWidgetBase):
    style = Style(prefix='torrentlist.ratio', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['ratio'].header),
                           style.attrs('header'))

    def get_mode(self):
        if 0 <= self.value < 1:
            return 'highlighted'

TUICOLUMNS['ratio'] = Ratio


class Size(_COLUMNS['size'], CellWidgetBase):
    style = Style(prefix='torrentlist.size', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['size'].header),
                           style.attrs('header'))

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['size'].header['right']

TUICOLUMNS['size'] = Size


class Downloaded(_COLUMNS['downloaded'], CellWidgetBase):
    style = Style(prefix='torrentlist.downloaded', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['downloaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        t = self.data
        if t['%downloaded'] < 100 and 0 < t['size-downloaded'] < t['size-final']:
            return 'highlighted'

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['downloaded'].header['right']

TUICOLUMNS['downloaded'] = Downloaded


class Uploaded(_COLUMNS['uploaded'], CellWidgetBase):
    style = Style(prefix='torrentlist.uploaded', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['uploaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        if self.data['size-uploaded'] < self.data['size-downloaded']:
            return 'highlighted'

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['uploaded'].header['right']

TUICOLUMNS['uploaded'] = Uploaded


class BytesAvailable(_COLUMNS['available'], CellWidgetBase):
    style = Style(prefix='torrentlist.available', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['available'].header),
                           style.attrs('header'))

    def get_mode(self):
        if self.value < self.data['size-final']:
            return 'highlighted'

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['available'].header['right']

TUICOLUMNS['available'] = BytesAvailable


class RateDown(_COLUMNS['rate-down'], CellWidgetBase):
    style = Style(prefix='torrentlist.rate-down', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-down'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else ''

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-down'].header['right']

TUICOLUMNS['rate-down'] = RateDown


class RateUp(_COLUMNS['rate-up'], CellWidgetBase):
    style = Style(prefix='torrentlist.rate-up', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-up'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else ''

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-up'].header['right']

TUICOLUMNS['rate-up'] = RateUp



class RateLimitDown(_COLUMNS['rate-limit-down'], CellWidgetBase):
    style = Style(prefix='torrentlist.rate-limit-down', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-limit-down'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < float('inf') else ''

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-limit-down'].header['right']

TUICOLUMNS['rate-limit-down'] = RateLimitDown


class RateLimitUp(_COLUMNS['rate-limit-up'], CellWidgetBase):
    style = Style(prefix='torrentlist.rate-limit-up', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-limit-up'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < float('inf') else ''

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-limit-up'].header['right']

TUICOLUMNS['rate-limit-up'] = RateLimitUp


class EtaComplete(_COLUMNS['eta'], CellWidgetBase):
    style = Style(prefix='torrentlist.eta', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['eta'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if bool(self.value) else ''

TUICOLUMNS['eta'] = EtaComplete


class TimeCreated(_COLUMNS['created'], CellWidgetBase):
    style = Style(prefix='torrentlist.created', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['created'].header),
                           style.attrs('header'))

TUICOLUMNS['created'] = TimeCreated

class TimeAdded(_COLUMNS['added'], CellWidgetBase):
    style = Style(prefix='torrentlist.added', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['added'].header),
                           style.attrs('header'))

TUICOLUMNS['added'] = TimeAdded

class TimeStarted(_COLUMNS['started'], CellWidgetBase):
    style = Style(prefix='torrentlist.started', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['started'].header),
                           style.attrs('header'))

TUICOLUMNS['started'] = TimeStarted

class TimeActive(_COLUMNS['activity'], CellWidgetBase):
    style = Style(prefix='torrentlist.activity', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['activity'].header),
                           style.attrs('header'))

TUICOLUMNS['activity'] = TimeActive

class TimeCompleted(_COLUMNS['completed'], CellWidgetBase):
    style = Style(prefix='torrentlist.completed', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['completed'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value.in_future else ''

TUICOLUMNS['completed'] = TimeCompleted


class Tracker(_COLUMNS['tracker'], CellWidgetBase):
    style = Style(prefix='torrentlist.tracker', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['tracker'].header),
                           style.attrs('header'))

TUICOLUMNS['tracker'] = Tracker


class Error(_COLUMNS['error'], CellWidgetBase):
    style = Style(prefix='torrentlist.error', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['error'].header),
                           style.attrs('header'))

TUICOLUMNS['error'] = Error


class Status(_COLUMNS['status'], CellWidgetBase):
    style = Style(prefix='torrentlist.status', focusable=True,
                  extras=('header',),
                  modes=('idle', 'downloading', 'uploading', 'connected', 'seeding',
                         'stopped', 'queued', 'isolated', 'verifying', 'discovering'))

    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['status'].header),
                           style.attrs('header'))

    MODE_MAP = {
        tkeys.Status.IDLE      : 'idle',
        tkeys.Status.DOWNLOAD  : 'downloading',
        tkeys.Status.UPLOAD    : 'uploading',
        tkeys.Status.CONNECTED : 'connected',
        tkeys.Status.SEED      : 'seeding',
        tkeys.Status.STOPPED   : 'stopped',
        tkeys.Status.QUEUED    : 'queued',
        tkeys.Status.ISOLATED  : 'isolated',
        tkeys.Status.VERIFY    : 'verifying',
        tkeys.Status.INIT      : 'discovering',
    }
    def get_mode(self):
        return self.MODE_MAP[self.value]

TUICOLUMNS['status'] = Status


class TorrentName(_COLUMNS['name'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='torrentlist.name', focusable=True, extras=('header',),
                  modes=('idle.progress1', 'idle.progress2', 'idle.complete',
                         'stopped.progress1', 'stopped.progress2', 'stopped.complete',
                         'isolated.progress1', 'isolated.progress2', 'isolated.complete',
                         'discovering.progress1', 'discovering.progress2', 'discovering.complete',
                         'verifying.progress1', 'verifying.progress2', 'verifying.complete',
                         'downloading.progress1', 'downloading.progress2', 'downloading.complete',
                         'uploading.progress1', 'uploading.progress2', 'uploading.complete',
                         'queued.progress1', 'queued.progress2', 'queued.complete',
                         'connected.progress1', 'connected.progress2', 'connected.complete'))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['name'].header),
                           style.attrs('header'))
    needed_keys = ('name', 'status', '%downloaded', '%verified', '%metadata',
                   'rate-up', 'rate-down', 'peers-connected')

    def __init__(self, *args, **kwargs):
        self.status = ('', 'idle', 0)
        super().__init__(*args, **kwargs)

    def update(self, torrent):
        new_status = self.make_status(torrent)
        if new_status != self.status:
            self.status = new_status
            self._invalidate()

    def render(self, size, focus=False):
        (maxcol,) = size
        name, mode, progress = self.status
        name = name[:maxcol].ljust(maxcol, ' ')  # Expand/Shrink name to full width
        if progress == 100:
            attrs = self.style.attrs(mode+'.complete', focused=focus)
            self.text.set_text((attrs, name))
        else:
            completed_col = int(maxcol * progress / 100)  # Width of first part of progress bar
            name1 = name[:completed_col]
            name2 = name[completed_col:]
            attrs1 = self.style.attrs(mode+'.progress1', focused=focus)
            attrs2 = self.style.attrs(mode+'.progress2', focused=focus)
            self.text.set_text([(attrs1, name1),
                                (attrs2, name2)])
        return super().render(size, focus)

    def get_mode(self):
        return self.status[1]

    @staticmethod
    def make_status(t):
        progress = t['%downloaded']
        Status = type(t['status'])
        if Status.STOPPED in t['status']:
            mode = 'stopped'
        elif Status.ISOLATED in t['status']:
            mode = 'isolated'
        elif Status.INIT in t['status']:
            mode = 'discovering'
            progress = t['%metadata']
        elif Status.VERIFY in t['status']:
            mode = 'verifying'
            progress = t['%verified']
        elif Status.DOWNLOAD in t['status']:
            mode = 'downloading'
        elif Status.UPLOAD in t['status']:
            mode = 'uploading'
        elif Status.QUEUED in t['status']:
            mode = 'queued'
        elif Status.CONNECTED in t['status']:
            mode = 'connected'
        else:
            mode = 'idle'
        return (t['name'], mode, progress)



TUICOLUMNS['name'] = TorrentName



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Value types for Torrent classes"""

# The TYPES dictionary at the end of this file maps Torrent key names to
# types.  Every Torrent key must have a type, even if it's just a no-op
# (`lambda obj: obj`).
#
# A type is any callable that converts a single value to the appropriate class
# instance.
#
# Types are used to convert values from the server (e.g. large integers) and
# from the user (e.g. number strings like '3.5G').  Not all types must accept
# user-given values (e.g. 'files').


from ..logging import make_logger
log = make_logger(__name__)

from collections import abc
from .utils import pretty_float
import os

from . import constants as const


from itertools import chain
import re
class Number(float):
    """float with a nice string representation; also parses strings like '123K' or '123Mi'"""

    _PREFIXES_BINARY = (('Ti', 1024**4), ('Gi', 1024**3), ('Mi', 1024**2), ('Ki', 1024))
    _PREFIXES_METRIC = (('T', 1000**4), ('G', 1000**3), ('M', 1000**2), ('k', 1000))
    _ALL_PREFIXES = tuple((prefix.lower(), size)
                          for prefix,size in chain.from_iterable(zip(_PREFIXES_BINARY,
                                                                     _PREFIXES_METRIC)))
    _ALL_PREFIXES_DCT = dict(_ALL_PREFIXES)
    _REGEX = re.compile('^([-+]?(?:\d+\.\d+|\d+|\.\d+)) ?(' +\
                        '|'.join(p[0] for p in _ALL_PREFIXES) + \
                        '|)(.*?)$',
                        flags=re.IGNORECASE)

    @classmethod
    def from_string(cls, string, prefix='metric', unit=None):
        match = cls._REGEX.match(string)
        if match is None:
            raise ValueError('Not a number: {!r}'.format(string))
        else:
            num = float(match.group(1))
            unit = match.group(3) or unit
            prfx = match.group(2)
            if prfx:
                all_prfxs = cls._ALL_PREFIXES_DCT
                prfx_lower = prfx.lower()
                if prfx_lower in all_prfxs:
                    num *= all_prfxs[prfx_lower]

            prfx_len = len(prfx)
            if prfx_len == 2:
                prefix = 'binary'
            elif prfx_len == 1:
                prefix = 'metric'

            return cls(num, prefix, unit)

    def __new__(cls, num, prefix='metric', unit=None):
        if isinstance(num, cls):
            return cls(float(num), prefix or num.prefix, unit or num.unit)

        obj = super().__new__(cls, num)
        if prefix == 'binary':
            obj._prefixes = cls._PREFIXES_BINARY
        elif prefix == 'metric':
            obj._prefixes = cls._PREFIXES_METRIC
        else:
            raise ValueError("prefix must be 'binary' or 'metric', not {!r}".format(prefix))
        obj.unit = unit
        obj.prefix = prefix
        return obj

    @property
    def with_unit(self):
        s = self.without_unit
        if self.unit is not None:
            s += self.unit
        return s

    @property
    def without_unit(self):
        for prefix,size in self._prefixes:
            if self >= size:
                return pretty_float(self/size) + prefix
        return pretty_float(self)

    def __str__(self):
        return self.without_unit

    def __repr__(self):
        return '<{} {}, prefix={!r}, unit={!r}>'.format(type(self).__name__, float(self),
                                                        self.prefix, self.unit)

    # Arithmetic operations return Number instances with unit and prefix preserved
    def __add__(self, other):
        return type(self)(super().__add__(other), unit=self.unit, prefix=self.prefix)
    def __sub__(self, other):
        return type(self)(super().__sub__(other), unit=self.unit, prefix=self.prefix)
    def __mul__(self, other):
        return type(self)(super().__mul__(other), unit=self.unit, prefix=self.prefix)
    def __div__(self, other):
        return type(self)(super().__div__(other), unit=self.unit, prefix=self.prefix)
    def __truediv__(self, other):
        return type(self)(super().__truediv__(other), unit=self.unit, prefix=self.prefix)
    def __floordiv__(self, other):
        return type(self)(super().__floordiv__(other), unit=self.unit, prefix=self.prefix)
    def __mod__(self, other):
        return type(self)(super().__mod__(other), unit=self.unit, prefix=self.prefix)
    def __divmod__(self, other):
        return type(self)(super().__divmod__(other), unit=self.unit, prefix=self.prefix)
    def __pow__(self, other):
        return type(self)(super().__pow__(other), unit=self.unit, prefix=self.prefix)

# Because 'convert' needs Number, which is specified in this file, it must be
# imported AFTER Number exists to avoid a circular import.
from . import convert


def _rate_limit(limit):
    return const.UNLIMITED if limit is None else convert.bandwidth(limit)


class Percent(float):
    """Float with a pretty string representation"""
    def __str__(self):
        return pretty_float(self)

def _calc_percent(a, b):
    try:
        return a / b * 100
    except ZeroDivisionError:
        return 0


class Ratio(Number):
    """A Torrent's upload/download ratio as a float"""
    UNKNOWN = -1
    NOT_APPLICABLE = -2
    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''
        else:
            return pretty_float(self)


class SeedCount(Number):
    UNKNOWN = -1
    def __str__(self):
        return '?' if self == self.UNKNOWN else super().__str__()


class Status(tuple):
    """A Torrent's status as string"""

    IDLE      = 'idle'
    DOWNLOAD  = 'downloading'
    UPLOAD    = 'uploading'
    CONNECTED = 'connected'
    SEED      = 'seeding'
    STOPPED   = 'stopped'
    QUEUED    = 'queued'
    ISOLATED  = 'isolated'
    VERIFY    = 'verifying'
    INIT      = 'discovering'
    ORDER = (VERIFY, DOWNLOAD, UPLOAD, INIT, CONNECTED,
             ISOLATED, QUEUED, IDLE, STOPPED, SEED)

    def __lt__(self, other):
        return self.ORDER.index(self[0]) < self.ORDER.index(other[0])

    def __le__(self, other):
        return self.ORDER.index(self[0]) <= self.ORDER.index(other[0])

    def __gt__(self, other):
        return self.ORDER.index(self[0]) > self.ORDER.index(other[0])

    def __ge__(self, other):
        return self.ORDER.index(self[0]) >= self.ORDER.index(other[0])



import operator
import unicodedata
class SmartCmpStr(str):
    """String with smart comparison capabilities

    Adds the <, >, <=, >= operators that compare length of strings and makes
    comparison case-insensitive if the other string consists solely of
    lower-case characters.
    """

    def __new__(cls, string):
        # Combine characters with diacritical marks ("a˚" -> "å") so len()
        # reports the correct length.
        # http://www.unicode.org/faq/char_combmark.html
        return super().__new__(cls, unicodedata.normalize('NFC', string))

    def __cmp(self, op, other):
        if not isinstance(other, str):
            return NotImplemented

        # Do case-insensitive comparison?
        # Make copies to avoid infinite recursion.
        o = str(other)
        if o == o.lower():
            s = str(self.lower())
        else:
            s = str(self)

        if op in (operator.__eq__, operator.__ne__, operator.__contains__):
            return op(s, o)
        elif self.isdigit():
            return op(int(s), len(o))
        elif other.isdigit():
            return op(len(s), int(o))
        else:
            return op(s, o)

    def __lt__(self, other): return self.__cmp(operator.lt, other)
    def __le__(self, other): return self.__cmp(operator.le, other)
    def __eq__(self, other): return self.__cmp(operator.eq, other)
    def __ne__(self, other): return self.__cmp(operator.ne, other)
    def __gt__(self, other): return self.__cmp(operator.gt, other)
    def __ge__(self, other): return self.__cmp(operator.ge, other)
    def __contains__(self, other): return self.__cmp(operator.contains, other)

    def __hash__(self):
        return super().__hash__()


class Path(SmartCmpStr):
    def __new__(cls, path):
        return super().__new__(cls, os.path.normpath(path))

    def __repr__(self):
        return '<{} {!r}>'.format(type(self).__name__, str(self))

    def __hash__(self):
        return super().__hash__()



SECONDS = (('y', 31557600),  # 365.25 days
           ('M',  2629800),  # 1y / 12
           ('d',    86400),
           ('h',     3600),
           ('m',       60),
           ('s',        1))

class Timedelta(int):
    # To sort unknown and not applicable Timedeltas below the rest, these
    # constants have large values that are very likely never encountered as
    # actual values.
    UNKNOWN        = 1e10    # ~3.1k years
    NOT_APPLICABLE = 1e10+1  # ~31k years

    _FROM_STRING_REGEX = re.compile((r'(\d+(?:\.\d+|)[' +
                                     r''.join(unit for unit,secs in SECONDS) +
                                     r']?)'), flags=re.IGNORECASE)
    @classmethod
    def from_string(cls, string):
        string = string.replace(' ', '')
        if len(string) < 1:
            raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, string))

        secs_total = 0
        for s in cls._FROM_STRING_REGEX.split(string):
            if len(s) < 1:
                continue
            elif not cls._FROM_STRING_REGEX.match(s):
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, s))
            elif s[-1].isdigit():
                # No unit specified
                secs_total += float(s)
            else:
                unit, num = s[-1], s[:-1]
                for unit_,secs in SECONDS:
                    if unit == unit_:
                        secs_total += float(num) * secs
                        break

        return cls(secs_total)

    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''
        elif self == 0:
            return 'now'

        abs_secs = abs(self)
        for i,(unit,amount) in enumerate(SECONDS):
            if abs_secs >= amount:
                num = self/amount

                # Small numbers get a sub-unit, for example '1d15h'
                if 1 <= abs_secs/amount < 10 and i < len(SECONDS)-1:
                    subunit, subamount = SECONDS[i+1]
                    if num >= 0:
                        subnum = abs( ((num%1) * amount) / subamount )
                    else:
                        subnum = abs( ((num%-1) * amount) / subamount )

                    if subnum >= 1:
                        return '%d%s%d%s' % (int(num), unit, int(subnum), subunit)

                return '%d%s' % (int(num), unit)

    @property
    def with_preposition(self):
        if self > 0:
            return 'in %s' % self
        elif self < 0:
            return ('%s ago' % self)[1:]  # Remove the first char ('-')
        else:
            return 'now'

    def __bool__(self):
        """Whether delta is known"""
        return self not in (self.UNKNOWN, self.NOT_APPLICABLE)

    @property
    def is_known(self):
        return bool(self)


import time
class Timestamp(int):
    UNKNOWN        = -1
    NOT_APPLICABLE = -2

    _FORMATS_DATE = (('%Y',       ('tm_year',)),
                     ('%Y-%m',    ('tm_year', 'tm_mon')),
                     ('%Y-%m-%d', ('tm_year', 'tm_mon', 'tm_mday')),
                     ('%d',       ('tm_mday',)),
                     ('%m-%d',    ('tm_mon', 'tm_mday')))
    _FORMATS_TIME = (('%H:%M', ('tm_hour', 'tm_min')),)

    # Create all combinations of date, time and date+time formats, keeping track
    # of the values they specify
    _FORMATS = []
    for date_frmt,date_given in _FORMATS_DATE:
        _FORMATS.append((date_frmt, date_given))
        for time_frmt,time_given in _FORMATS_TIME:
            _FORMATS.append((time_frmt, time_given))
            given = date_given + time_given
            _FORMATS.append(('%s %s' % (date_frmt, time_frmt), given))
            _FORMATS.append(('%s %s' % (time_frmt, date_frmt), given))

    @classmethod
    def from_string(cls, string):
        string = string.strip().replace('  ', ' ')

        def fill_in_missing_values(t, given):
            # Today with seconds set to 0
            t_now = time.localtime()
            t_now = time.struct_time((t_now.tm_year, t_now.tm_mon, t_now.tm_mday,
                                      t_now.tm_hour, t_now.tm_min, 0,
                                      t_now.tm_wday, t_now.tm_yday, -1))

            # THISYEAR-01-01 00:00:00
            t_default = time.struct_time((t_now.tm_year, 1, 1, 0, 0, 0, 0, 1, -1))

            names = ('tm_year', 'tm_mon', 'tm_mday', 'tm_hour', 'tm_min',
                     'tm_sec', 'tm_wday', 'tm_yday', 'tm_isdst')
            args = []

            # Copy values from `t` if they are in `given`, otherwise from
            # `t_now` or `t_default`.
            for name in names:
                if name in given:
                    args.append(getattr(t, name))
                else:
                    if names.index(given[0]) > names.index(name):
                        args.append(getattr(t_now, name))
                    else:
                        args.append(getattr(t_default, name))
            return time.struct_time(args)

        t = None
        for frmt, given in cls._FORMATS:
            try:
                t = time.strptime(string, frmt)
            except ValueError:
                pass
            else:
                t = fill_in_missing_values(t, given)
                break

        if t is None:
            raise ValueError('Invalid format: %r' % string)
        else:
            return cls(time.mktime(t))

    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''

        abs_delta = abs(self - time.time())
        if abs_delta <= SECONDS[2][1]:  # <= 1 day
            frmt = '%H:%M'
        else:
            frmt = '%Y-%m-%d'
        return time.strftime(frmt, time.localtime(self))

    @property
    def full(self):
        if self == self.UNKNOWN:
            return 'unknown'
        elif self == self.NOT_APPLICABLE:
            return 'not applicable'
        else:
            return time.strftime('%Y-%m-%d %H:%M', time.localtime(self))

    def __bool__(self):
        """Whether timestamp known"""
        return self != self.UNKNOWN and self != self.NOT_APPLICABLE

    @property
    def is_known(self):
        return bool(self)

    @property
    def delta(self):
        if self == self.UNKNOWN:
            return Timedelta(Timedelta.UNKNOWN)
        elif self == self.NOT_APPLICABLE:
            return Timedelta(Timedelta.NOT_APPLICABLE)
        else:
            return Timedelta(self - time.time())

    @property
    def in_future(self):
        return bool(self) and self > time.time()



from functools import total_ordering
@total_ordering
class TorrentFilePriority(str):
    INT2STR = {-1:'low', 0:'normal', 1:'high', -2:'shun'}
    STR2INT = {'low':-1, 'normal':0, 'high':1, 'shun':-2}

    def __new__(cls, prio):
        if isinstance(prio, int):
            if prio not in cls.INT2STR:
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, prio))
            obj = super().__new__(cls, cls.INT2STR[prio])
        else:
            if prio not in cls.STR2INT:
                raise ValueError('Invalid {} value: {!r}'.format(cls.__name__, prio))
            obj = super().__new__(cls, prio)
        return obj

    def __int__(self): return self.STR2INT[self]
    def __lt__(self, other): return int(self) < int(other)
    def __repr__(self): return '<%s %r>' % (type(self).__name__, str(self))


import os
class TorrentFile(abc.Mapping):
    """Mapping that holds the values of a single file in a torrent"""

    # Distinguish subtrees from files without comparing classes everywhere
    nodetype = 'leaf'

    TYPES = {
        'tid'             : int,
        'id'              : int,
        'name'            : SmartCmpStr,
        'path'            : Path,
        'size-total'      : convert.size,
        'size-downloaded' : convert.size,
        'is-wanted'       : bool,
        'priority'        : TorrentFilePriority,
        'progress'        : Percent,
    }

    _VALUES = {
        'tid'             : lambda raw: raw['tid'],
        'id'              : lambda raw: raw['id'],
        'name'            : lambda raw: raw['name'],
        'path'            : lambda raw: os.sep.join(raw['path']),
        'size-total'      : lambda raw: raw['size-total'],
        'size-downloaded' : lambda raw: raw['size-downloaded'],
        'is-wanted'       : lambda raw: raw['is-wanted'],
        'priority'        : lambda raw: (TorrentFilePriority.STR2INT['shun']
                                         if not raw['is-wanted'] else raw['priority']),
        'progress'        : lambda raw: _calc_percent(raw['size-downloaded'], raw['size-total']),
    }

    def __init__(self, tid, id, name, path, size_total, size_downloaded, is_wanted, priority):
        self._raw = {'tid': tid, 'id': id, 'name': name, 'path': path,
                     'is-wanted': is_wanted, 'priority': priority,
                     'size-total': size_total, 'size-downloaded': size_downloaded}
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            self._cache[key] = self.TYPES[key](
                self._VALUES[key](self._raw)
            )
        return self._cache[key]

    def update(self, raw):
        self._raw.update(raw)
        try:
            for k in raw:
                del self._cache[k]
            if 'size-downloaded' in raw:
                del self._cache['progress']
        except KeyError:
            pass

    def __repr__(self): return '<{} {!r}>'.format(type(self).__name__, self['name'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)


from . import base
def _ensure_TorrentFileTree(obj):
    if isinstance(obj, base.TorrentFileTreeBase):
        return obj
    else:
        raise RuntimeError('Not a TorrentFileTreeBase instance: {!r}'.format(obj))




import time
from collections import (defaultdict, deque)
MAX_SAMPLES = 10
MAX_SAMPLE_AGE = 5*3600
_PEER_PROGRESS_DATA = defaultdict(lambda: deque(maxlen=MAX_SAMPLES))

def gc_peer_progress_data():
    for peer_id,samples in tuple(_PEER_PROGRESS_DATA.items()):
        # Remove samples that are too old
        while samples and (samples[0][0] + MAX_SAMPLE_AGE) < time.monotonic():
            samples.popleft()

        # Remove deque if there are no samples left
        if not samples:
            del _PEER_PROGRESS_DATA[peer_id]

def _guess_peer_rate_and_eta(peer_id, peer_progress, torrent_size):
    rate = 0
    if peer_progress >= 1:
        # Peer has already downloaded everything
        eta = Timedelta.NOT_APPLICABLE
    else:
        eta = Timedelta.UNKNOWN
        samples = _PEER_PROGRESS_DATA[peer_id]

        # Don't add the same progress twice
        if not samples or peer_progress != samples[-1][1]:
            samples.append((time.monotonic(), peer_progress))

        # We need at least 3 samples
        if len(samples) >= 3:
            # Use second and last sample to calculate rate.  The first sample is
            # ignored because its timestamp may be inaccurate: When we add the
            # first sample, the peer's progress is not current but the latest we
            # received, which happened likely tens of seconds ago.
            t_first, p_first = samples[1]
            t_last, p_last = samples[-1]
            p_diff = p_last - p_first
            t_diff = t_last - t_first

            # It's possible progress goes down, e.g. if a peer deletes a file
            if p_diff > 0:
                size_diff = torrent_size * p_diff
                rate = size_diff / t_diff
                size_remaining = torrent_size - (torrent_size * peer_progress)
                eta = size_remaining / rate

    return rate, eta

from . import geoip
class TorrentPeer(abc.Mapping):
    TYPES = {
        'id'        : lambda val: val,
        'tid'       : lambda val: val,
        'tname'     : SmartCmpStr,
        'tsize'     : convert.size,
        'ip'        : str,
        'port'      : int,
        'client'    : SmartCmpStr,
        'country'   : SmartCmpStr,
        'progress'  : Percent,
        'rate-up'   : convert.bandwidth,
        'rate-down' : convert.bandwidth,
        'eta'       : Timedelta,
        'rate-est'  : convert.bandwidth,
    }

    _VALUES = {
        'id'      : lambda p: hash((p['tid'], p['ip'], p['port'])),
        'country' : lambda p: geoip.country_code(p['ip']) or '?',
    }

    def __init__(self, tid, tname, tsize, ip, port, client, progress, rate_up, rate_down):
        self._dct = {'tid': tid, 'tname': tname, 'tsize': tsize,
                     'ip': ip, 'port': port, 'client': client, 'progress': progress,
                     'rate-up': rate_up, 'rate-down': rate_down}
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            if key in ('eta', 'rate-est'):
                rate, eta = _guess_peer_rate_and_eta(self['id'], self['progress']/100, self['tsize'])
                self._cache['rate-est'] = self.TYPES['rate-est'](rate)
                self._cache['eta'] = self.TYPES['eta'](eta)

            else:
                if key in self._VALUES:
                    val = self._VALUES[key](self._dct)
                else:
                    val = self._dct[key]
                self._cache[key] = self.TYPES[key](val)
        return self._cache[key]

    def __repr__(self): return '<{} {}@{}>'.format(type(self).__name__, self['ip'], self['progress'])
    def __iter__(self): return iter(self.TYPES)
    def __len__(self): return len(self.TYPES)



TYPES = {
    'id'                : int,
    'hash'              : str,
    'name'              : SmartCmpStr,
    'ratio'             : Ratio,
    'status'            : Status,
    'path'              : Path,
    'private'           : bool,
    'comment'           : SmartCmpStr,
    'creator'           : SmartCmpStr,
    'magnetlink'        : str,
    'count-pieces'      : Number,

    '%downloaded'       : Percent,
    '%uploaded'         : Percent,
    '%metadata'         : Percent,
    '%verified'         : Percent,
    '%available'        : Percent,

    'peers-connected'   : Number,
    'peers-uploading'   : Number,
    'peers-downloading' : Number,
    'peers-seeding'     : SeedCount,

    'timespan-eta'      : Timedelta,
    'time-created'      : Timestamp,
    'time-added'        : Timestamp,
    'time-started'      : Timestamp,
    'time-activity'     : Timestamp,
    'time-completed'    : Timestamp,
    'time-manual-announce-allowed': Timestamp,

    'rate-down'         : convert.bandwidth,
    'rate-up'           : convert.bandwidth,

    'rate-limit-down'   : _rate_limit,
    'rate-limit-up'     : _rate_limit,

    'size-final'        : convert.size,
    'size-total'        : convert.size,
    'size-downloaded'   : convert.size,
    'size-uploaded'     : convert.size,
    'size-available'    : convert.size,
    'size-left'         : convert.size,
    'size-corrupt'      : convert.size,
    'size-piece'        : convert.size,

    'trackers'          : tuple,
    'error'             : str,
    'peers'             : tuple,
    'files'             : _ensure_TorrentFileTree,
}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

import urwid
from ..scroll import (ScrollBar, Scrollable)


def mksection(title, width, items):
    # Setting class variable 'title = title' below produces "NameError: name
    # 'title' is not defined"
    title_, width_ = title, width
    class Section(urwid.WidgetWrap):
        title = title_
        width = width_

        def __init__(self):
            value_widgets = {}
            needed_keys = set()
            rows = []
            label_width = max(len(item.label) for item in items)
            for item in items:
                label_w = urwid.Text(item.label.rjust(label_width))
                value_w = urwid.Text('')
                value_widgets[item] = value_w
                rows.append(urwid.Columns([('pack', label_w),
                                           ('pack', urwid.Text(': ')),
                                           value_w]))
                needed_keys.update(item.needed_keys)
            self._value_widgets = value_widgets
            self.needed_keys = needed_keys
            super().__init__(urwid.Pile(rows))

        def update(self, torrent):
            for item,value_w in self._value_widgets.items():
                value_w.set_text(item.human_readable(torrent))

    return Section


_sections = []
from ...views.tdetails import SECTIONS
for section in SECTIONS:
    sectionw = mksection(**section)
    _sections.append(sectionw)


class TorrentDetailsWidget(urwid.WidgetWrap):
    def __init__(self, srvapi, tid, title=None):
        self._title = title
        self._tid = None

        sections = []
        self._sections = {}
        for section_cls in _sections:
            section = section_cls()
            sections.append(section)
            self._sections[section.title] = section

        def add_title(title, section):
            header = urwid.Columns([('pack', urwid.Text('──┤ %s ├' % title)),
                                    urwid.Divider('─')])
            return urwid.Pile([('pack', header), section])

        grid = urwid.GridFlow([], cell_width=1, h_sep=3, v_sep=1, align='left')
        for section in sections:
            opts = grid.options('given', section.width)
            section_wrapped = add_title(section.title, section)
            grid.contents.append((section_wrapped, opts))

        grid_sb = urwid.AttrMap(
            ScrollBar(urwid.AttrMap(Scrollable(grid), 'torrentdetails')),
            'scrollbar'
        )
        super().__init__(grid_sb)

        # Register new request in request pool
        keys = set(('name',)).union(key for w in sections for key in w.needed_keys)
        self._poller = srvapi.create_poller(srvapi.torrent.torrents, (tid,), keys=keys)
        self._poller.on_response(self._handle_response)

    def _handle_response(self, response):
        if response is not None and response.success:
            torrent = response.torrents[0]
            self._tid = torrent['id']
            for w in self._sections.values():
                w.update(torrent)

            # Set new tab title if necessary
            if self._title is None:
                self._title = torrent['name']
            if self.title_updater is not None:
                self.title_updater(self._title)
        else:
            self._tid = None

    @property
    def title(self):
        return self._title or ''

    @property
    def focused_torrent_id(self):
        return self._tid



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Converters for units"""

from ..logging import make_logger
log = make_logger(__name__)

from .tkeys import Number


class _DataCountConverter():
    """Convert bits or bytes with metric or binary unit prefix to Number"""

    def __init__(self):
        self.prefix = 'metric'
        self.unit = 'byte'

    # Units (Long2Short and Short2Long)
    UNITS_L2S = {'bit': 'b', 'byte': 'B', 'b': 'b', 'B': 'B'}
    UNITS_S2L = {'b': 'bit', 'B': 'byte', 'bit': 'bit', 'byte': 'byte'}

    def from_string(self, string):
        """Parse number of bytes/bits from `string`

        `string` must be of the format "NUMBER[UNIT PREFIX][UNIT]".  UNIT PREFIX
        is something like 'M' (mega) or 'Mi' (mebi). UNIT must be 'b' (bits) or
        'B' (bytes). If UNIT is not given, NUMBER is assumed to be in whatever
        the `unit` property is set to.

        Raises ValueError if unit is invalid.
        """
        num = Number.from_string(string)
        unit = num.unit
        if unit is None:
            # Default to value of `unit` property
            unit = self.unit
            num.unit = self.UNITS_L2S[unit]
        elif unit not in self.UNITS_S2L:
            raise ValueError("Unit must be 'b' (bit) or 'B' (byte), not {!r}".format(unit))

        return self._ensure_unit(num)

    def __call__(self, bytes):
        """Make Number from `bytes`

        The returned Number object is converted to bits if the `unit` property
        is set to 'bits'.

        If `bytes` is a Number object, it's `unit` property is evaluated to
        ensure the correct unit of the returned Number.
        """
        if isinstance(bytes, Number):
            return self._ensure_unit(bytes)
        elif self._unit == 'bit':
            return Number(bytes*8, prefix=self._prefix, unit='b')
        else:
            return Number(bytes, prefix=self._prefix, unit='B')

    def _ensure_unit(self, num):
        # Make sure the unit of the returned Number is what the `unit` property is set to
        unit_wanted = self._unit
        unit_given = self.UNITS_S2L[num.unit]
        if unit_wanted == unit_given:
            return num
        elif unit_given == 'bit':
            return Number(num/8, prefix=self._prefix, unit='B')
        elif unit_given == 'byte':
            return Number(num*8, prefix=self._prefix, unit='b')

    @property
    def unit(self):
        """'bit' or 'byte'"""
        return self._unit

    @unit.setter
    def unit(self, unit):
        if unit not in ('bit', 'byte'):
            raise ValueError("unit must be 'bit' or 'byte'")
        self._unit = unit

    @property
    def prefix(self):
        """'binary' or 'metric'"""
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        if prefix not in ('binary', 'metric'):
            raise ValueError("unit must be 'binary' or 'metric'")
        else:
            self._prefix = prefix


bandwidth = _DataCountConverter()

class bandwidth_mixin():
    @property
    def bandwidth_unit(self):
        """Bandwidth numbers are displayed in 'bit' or 'byte'"""
        return bandwidth.unit

    @bandwidth_unit.setter
    def bandwidth_unit(self, unit):
        bandwidth.unit = unit

    @property
    def bandwidth_prefix(self):
        """Bandwidth numbers are displayed with 'binary' or 'metric' unit prefixes"""
        return bandwidth.prefix

    @bandwidth_prefix.setter
    def bandwidth_prefix(self, prefix):
        bandwidth.prefix = prefix


size = _DataCountConverter()

class size_mixin():
    @property
    def size_unit(self):
        """Size numbers are displayed in 'bit' or 'byte'"""
        return size.unit

    @size_unit.setter
    def size_unit(self, unit):
        size.unit = unit

    @property
    def size_prefix(self):
        """Size numbers are displayed with 'binary' or 'metric' unit prefixes"""
        return size.prefix

    @size_prefix.setter
    def size_prefix(self, prefix):
        size.prefix = prefix



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""General hooks that are always needed regardless of the interface"""

from .main import (cfg, srvapi)
from .views.tlist import COLUMNS as TCOLUMNS
from .views.flist import COLUMNS as FCOLUMNS
from .views.plist import COLUMNS as PCOLUMNS


def _set_rpc_timeout(timeout):
    srvapi.rpc.timeout = timeout.value
cfg['srv.timeout'].on_change(_set_rpc_timeout)


def _set_rpc_url(url):
    srvapi.url = url.value
cfg['srv.url'].on_change(_set_rpc_url)


def _set_bandwidth_unit(unit):
    srvapi.bandwidth_unit = unit.value
    u = {'bit': 'b', 'byte': 'B'}[unit.value]
    TCOLUMNS['rate-up'].set_unit(u)
    TCOLUMNS['rate-down'].set_unit(u)
    TCOLUMNS['rate-limit-up'].set_unit(u)
    TCOLUMNS['rate-limit-down'].set_unit(u)

    PCOLUMNS['rate-up'].set_unit(u)
    PCOLUMNS['rate-down'].set_unit(u)
    PCOLUMNS['rate-est'].set_unit(u)
cfg['unit.bandwidth'].on_change(_set_bandwidth_unit)
_set_bandwidth_unit(cfg['unit.bandwidth'])  # Initially call TCOLUMNS[...].set_unit()


def _set_bandwidth_prefix(prefix):
    srvapi.bandwidth_prefix = prefix.value
cfg['unitprefix.bandwidth'].on_change(_set_bandwidth_prefix)


def _set_size_unit(unit):
    srvapi.size_unit = unit.value
    u = {'bit': 'b', 'byte': 'B'}[unit.value]
    TCOLUMNS['size'].set_unit(u)
    TCOLUMNS['downloaded'].set_unit(u)
    TCOLUMNS['uploaded'].set_unit(u)
    TCOLUMNS['available'].set_unit(u)

    FCOLUMNS['size'].set_unit(u)
    FCOLUMNS['downloaded'].set_unit(u)
cfg['unit.size'].on_change(_set_size_unit)
_set_size_unit(cfg['unit.size'])  # Initially call TCOLUMNS[...].set_unit()


def _set_size_prefix(prefix):
    srvapi.size_prefix = prefix.value
cfg['unitprefix.size'].on_change(_set_size_prefix)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...main import cfg
from .tlist_columns import TUICOLUMNS as TCOLUMNS
from .flist_columns import TUICOLUMNS as FCOLUMNS
from .plist_columns import TUICOLUMNS as PCOLUMNS


def _set_bandwidth_unit(unit):
    u = {'bit': 'b', 'byte': 'B'}[unit.value]
    TCOLUMNS['rate-up'].set_unit(u)
    TCOLUMNS['rate-down'].set_unit(u)
    TCOLUMNS['rate-limit-up'].set_unit(u)
    TCOLUMNS['rate-limit-down'].set_unit(u)

    PCOLUMNS['rate-up'].set_unit(u)
    PCOLUMNS['rate-down'].set_unit(u)
    PCOLUMNS['rate-est'].set_unit(u)
cfg['unit.bandwidth'].on_change(_set_bandwidth_unit)


def _set_size_unit(unit):
    u = {'bit': 'b', 'byte': 'B'}[unit.value]
    TCOLUMNS['size'].set_unit(u)
    TCOLUMNS['downloaded'].set_unit(u)
    TCOLUMNS['uploaded'].set_unit(u)
    TCOLUMNS['available'].set_unit(u)

    FCOLUMNS['size'].set_unit(u)
    FCOLUMNS['downloaded'].set_unit(u)
cfg['unit.size'].on_change(_set_size_unit)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from collections import OrderedDict
DESCRIPTIONS = OrderedDict()

# This is a function so all the objects get garbage collected after
# parsing finished
def parse():
    import argparse
    _parser = argparse.ArgumentParser(add_help=False)
    def _add_arg(*args, section='OPTIONS', description=None, varname=None, **kwargs):
        if description is not None:
            optstr = ','.join(args)
            if varname is not None:
                optstr += ' %s' % varname

            if section not in DESCRIPTIONS:
                DESCRIPTIONS[section] = OrderedDict()
            DESCRIPTIONS[section][optstr] = description
        _parser.add_argument(*args, **kwargs)


    _add_arg('--help','-h', nargs='*', default=None,
             section='OPTIONS',
             description="Display help about TOPIC",
             varname='[TOPIC]')

    _add_arg('--version','-v', action='store_true',
             section='OPTIONS',
             description="Display version number and exit")


    _add_arg('--tui', '-t', action='store_true',
             section='OPTIONS',
             description='Enforce the TUI')
    _add_arg('--notui', '--no-tui', '-T', action='store_true',
             section='OPTIONS',
             description='Inhibit the TUI')


    _add_arg('--rcfile', '--rc-file', '-c',
             section='OPTIONS',
             description='Run commands from FILE upon startup',
             varname='FILE')
    _add_arg('--norcfile', '--no-rc-file', '-C', action='store_true',
             section='OPTIONS',
             description='Do not run commands from any rc file')


    _add_arg('--debug', type=lambda mods: mods.split(','), default=[],
             section='DEVELOPER OPTIONS',
             description=('Log debug messages from comma-separated list of MODULES'
                          ' (e.g. "client,commands.tui")'),
             varname='MODULES')
    _add_arg('--debug-file', default=None,
             section='DEVELOPER OPTIONS',
             description='Log debug messages to FILE',
             varname='FILE')
    _add_arg('--profile-file', default=None,
             section='DEVELOPER OPTIONS',
             description='Write cProfile statistics to FILE',
             varname='FILE')

    # Anything not specified above is a subcommand or a subcommand option.
    _parser.add_argument('subcmds', nargs=argparse.REMAINDER)
    args = vars(_parser.parse_args())
    _subcmds = args.pop('subcmds')

    # Convert -h option to 'help' command
    if args['help'] is not None:
        _subcmds.append('help')
        _subcmds.extend(args['help'])

    # Convert -v option to 'version' command
    if args['version']:
        _subcmds.append('version')


    # Assemble commands into a command sequence (see CommandManager.split_cmdchain)
    from .commands import is_op
    _cmds = [[]]
    for arg in _subcmds:
        if is_op(arg):
            _cmds.append(arg)
            _cmds.append([])       # Start new command
        else:
            _cmds[-1].append(arg)  # Append arg to current command
    cmds = [cmd for cmd in _cmds if cmd]

    return args,cmds



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import weakref

from . import convert
from .utils import SleepUneasy

from .aiotransmission.rpc import TransmissionRPC
from .aiotransmission.api_status import StatusAPI
from .aiotransmission.api_settings import SettingsAPI
from .aiotransmission.api_torrent import TorrentAPI

from .poll import RequestPoller
from .trequestpool import TorrentRequestPool


# https://stackoverflow.com/a/6849299
class _lazy_property():
    """Property that replaces itself with the requested object"""

    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        value = self.fget(obj)
        setattr(obj, self.func_name, value)
        setattr(obj, self.func_name + '_created', True)
        obj.manage_pollers_now()
        return value


class API(convert.bandwidth_mixin, convert.size_mixin):
    """Provide and manage *API classes as singletons

    A convenience class that creates instances of TransmissionRPC, TorrentAPI,
    StatusAPI, TorrentRequestPool and TorrentCounters in a lazy manner on
    demand.
    """
    def __init__(self, url, interval=1, loop=None):
        if loop is None:
            raise TypeError('Missing argument: loop')
        self.loop = loop
        self._url = url
        self._interval = interval
        self._pollers = []
        self._manage_pollers_interval = SleepUneasy(loop=self.loop)

    @property
    def url(self):
        """URL to API interface"""
        if self.created('rpc'):
            return self.rpc.url
        return self._url

    @url.setter
    def url(self, url):
        self._url = url
        if self.created('rpc'):
            self.rpc.url = url

    def created(self, prop):
        """Whether property `prop` was created"""
        return hasattr(self, prop+'_created')

    @_lazy_property
    def rpc(self):
        """TransmissionRPC singleton"""
        log.debug('Creating RPC singleton')
        return TransmissionRPC(self._url, loop=self.loop)

    @_lazy_property
    def torrent(self):
        """TorrentAPI singleton"""
        log.debug('Creating TorrentAPI singleton')
        return TorrentAPI(self.rpc)

    @_lazy_property
    def status(self):
        """StatusAPI singleton"""
        log.debug('Creating StatusAPI singleton')
        return StatusAPI(self, interval=self._interval)

    @_lazy_property
    def settings(self):
        """SettingsAPI singleton"""
        log.debug('Creating SettingsAPI singleton')
        return SettingsAPI(self, interval=self._interval)

    @_lazy_property
    def treqpool(self):
        """TorrentRequestPool singleton"""
        log.debug('Creating TorrentRequestPool singleton')
        return TorrentRequestPool(self, interval=self._interval)

    def create_poller(self, *args, interval=None, loop=None, **kwargs):
        """Create, start and return new RequestPoller instance

        All arguments are used to create the poller, except for `interval` and
        `loop`, which are ignored and replaced with this object's `interval`
        and `loop` attributes.

        The RequestPoller instance is treated like all other pollers, i.e. it
        is polled when `poll` is called, its interval is changed when
        `interval` is set, etc.
        """
        poller = RequestPoller(*args, interval=self.interval, loop=self.loop, **kwargs)
        self._pollers.append(poller)
        self.manage_pollers_now()
        return poller

    async def _manage_pollers(self):
        def is_needed(poller):
            # Whether anyone is still interested in the poller
            try:
                return poller.has_callbacks
            except AttributeError:
                try:
                    return poller.has_subscribers
                except AttributeError:
                    return True

        async def manage():
            for poller in self._existing_pollers:
                running = poller.running
                needed = is_needed(poller)

                if not running and needed:
                    log.debug('Starting because not running and has callbacks: %r', poller)
                    await poller.start()
                elif running and not needed:
                    log.debug('Stopping because running and no callbacks: %r', poller)
                    await poller.stop()
                    if poller in self._pollers:
                        self._pollers.remove(poller)

        # To estimate peer download rates, we have to keep track of all the
        # peers' progresses, and we need to prune that data from time to
        # time. It's a bit dirty to put this here, but managing another task
        # just for this isn't much better.
        from .tkeys import gc_peer_progress_data
        while True:
            gc_peer_progress_data()

            log.debug('Managing pollers:')
            await manage()

            # If a poller was added while we were managing the existing
            # pollers, it won't get started until the next interval.  This
            # also solves a weird bug that accumulates unused pollers when
            # rapidly opening and closing file lists.
            while any(not poller.running for poller in self._pollers):
                log.debug('Managing pollers again:')
                await manage()

            await self._manage_pollers_interval.sleep(10)

    def manage_pollers_now(self):
        self._manage_pollers_interval.interrupt()

    # Standard pollers accessible through properties
    _STD_POLLERS = ('status', 'settings', 'treqpool')
    @property
    def _existing_pollers(self):
        for pname in self._STD_POLLERS:
            if self.created(pname):
                yield getattr(self, pname)
        yield from self._pollers

    @property
    def interval(self):
        """Delay between polls of all pollers"""
        return self._interval

    @interval.setter
    def interval(self, interval):
        self._interval = interval
        for poller in self._existing_pollers:
            poller.interval = interval

    @property
    def pollers_running(self):
        """Whether pollers are running or not"""
        pollers = tuple(self._existing_pollers)
        if not pollers:
            return False
        return pollers[0].running

    def poll(self):
        """Poll all created and running pollers immediately

        This also resets the interval - the next polls are made `interval`
        seconds later.
        """
        for poller in self._existing_pollers:
            if poller.running:
                poller.poll()

    async def start_polling(self):
        """Start all created pollers"""
        for poller in self._existing_pollers:
            if not poller.running:
                await poller.start()
        self._manage_pollers_task = self.loop.create_task(self._manage_pollers())
        self._manage_pollers_task.add_done_callback(lambda task: task.result())

    async def stop_polling(self):
        """Stop all created pollers"""
        for poller in self._existing_pollers:
            if poller.running:
                await poller.stop()
        self._manage_pollers_task.cancel()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Anything related to the help system that is common between interfaces"""

from .logging import make_logger
log = make_logger(__name__)

from collections import abc
from . import (APPNAME, __version__)
from .utils import (striplines, expandtabs)

from .cliopts import DESCRIPTIONS

MAIN_TOPICS = ('commands', 'settings', 'keymap', 'filter', 'rcfile')

ALIASES = {
    'command': 'commands', 'cmds': 'commands', 'cmd': 'commands',
    'setting': 'settings', 'config': 'settings', 'cfg': 'settings',
    'keys': 'keymap', 'keybindings': 'keymap',
    'filters': 'filter', 'filtering': 'filter',
}


def finalize_lines(lines):
    return tuple(
        striplines(line.format(APPNAME=APPNAME)
                   for line in expandtabs.expand(lines))
    )


class HelpManager():
    """Provide help texts for CommandManager, Settings and KeyMap objects"""
    def __init__(self):
        self._cmdmgr = None
        self._settings = None
        self._keymap = None

    def find(self, topic=None):
        """Find help topic and return lines"""
        if topic in ALIASES:
            topic = ALIASES[topic]

        if topic is None:
            return self.overview
        elif topic in MAIN_TOPICS:
            return getattr(self, topic)
        elif topic in self._cmdmgr:
            return self.command(topic)
        elif topic in self._settings:
            return self.setting(topic)
        else:
            raise ValueError('Unknown help topic: {!r}'.format(topic))

    @property
    def overview(self):
        lines = [
            '{} {}'.format(APPNAME, __version__),
            '',
            'SYNTAX',
            '\t{APPNAME} [<OPTIONS>] [<COMMAND> ; <COMMAND> ; <COMMAND>...]',
            '',
        ]

        for section,opts in DESCRIPTIONS.items():
            lines.append('%s' % section.upper())
            for opts,desc in opts.items():
                lines.append('\t%s  \t%s' % (opts, desc))
            lines.append('')

        lines.append('For more information run:')
        for topic in MAIN_TOPICS:
            lines.append('\thelp %s' % topic)
        return finalize_lines(lines)

    @property
    def settings(self):
        """Must be set to a Settings object; provides a help text"""
        cfg = self._settings

        lines = [
            'SETTINGS',
            "\tSettings can be changed with the commands 'set' and 'reset'.",
            '',
            "\tUse an rc file (see 'help rcfile') to specify your custom defaults.",
            '',
        ]

        if cfg is None:
            lines.append('\tSettings have not been loaded.')
        else:
            lines.append('\tAVAILABLE SETTINGS')
            for name in sorted(cfg.names):
                lines.append('\t\t' + name + '  \t' + cfg[name].description)
        return finalize_lines(lines)

    @settings.setter
    def settings(self, settings):
        self._settings = settings

    def setting(self, name):
        """Return help text for setting"""
        v = self._settings[name]
        lines = [v.name,
                 '\t' + v.description,
                 '\tType: \t' + v.typename,
                 '\tValue: \t' + v.str(default=False),
                 '\tDefault: \t' + v.str(default=True)]
        if hasattr(v, 'options'):
            lines.append('\tOptions: \t' + \
                         ', '.join(str(o) for o in sorted(v.options)))
        if hasattr(v, 'valuesyntax'):
            lines.append('\tSyntax: \t{}'.format(v.valuesyntax))
        return finalize_lines(lines)

    @property
    def commands(self):
        """Must be set to a CommandManager object; provides a help text"""
        lines = [
            'COMMANDS',
            '\tCommands can be called ',
            '\t\t- \tby providing them as command line arguments,',
            "\t\t- \tvia the internal command line (hit ':' to open it),",
            "\t\t- \tby binding them to keys (see 'help bind'),",
            ("\t\t- \tby listing them in an rc file (see 'help rcfile') "
             "and loading it with the '--rcfile' option or the 'rc' command."),
            '',
            '\tCHAINING COMMANDS',
            ("\t\tCombining commands with operators makes it possible to run "
             "a command based on the previous command's success."),
             "",
             "\t\tAvailable command operators are: ",
             "\t\t\t&/and \t- \tRun the next command if the previous command succeeded.",
             "\t\t\t|/or \t- \tRun the next command if the previous command failed.",
             "\t\t\t;/also \t- \tRun the next command in any case.",
             "",
             "\t\tCommand operators must be enclosed by spaces.",
             "",
            ("\t\tFor example, 'ls foo & ls bar' would list all 'foo' torrents and, "
             "if any where found, continue to list all 'bar' torrents.  "
             "However, 'ls foo | ls bar' would list 'bar' torrents only if there "
             "are no 'foo' torrents."),
            '',
            '\tGUESSING THE USER INTERFACE (CLI/TUI)',
            ("\t\tIf commands are given as command line arguments and neither "
             "'--tui' nor '--notui' are provided, {APPNAME} tries to guess "
             "whether it makes sense to start the TUI or just run the commands "
             "and exit.  For example, if you run '{APPNAME} stop foo', "
             "it is reasonable to assume that you want to run 'stop foo' and "
             "get your shell prompt back.  But if you run '{APPNAME} set srv.url foo.bar', "
             "you probably expect the TUI to pop up."),
             '',
             "\t\tThis is how this works basically:",
             ("\t\t\t- \tWithout CLI commands, the TUI is loaded."),
             ("\t\t\t- \tCommands in the TORRENT category (see below) prevent the TUI."),
             ("\t\t\t- \tChanging TUI settings ('(re)set tui.*') enforces the TUI."),
             ("\t\t\t- \tChanging server settings ('(re)set srv.*') prevents the TUI."),
             ("\t\t\t- \tCommands that are exclusive to TUI or CLI "
              "(e.g. 'tab') enforce their interface.  Providing both TUI- "
              "and CLI-only commands produces an error."),
            '',
        ]

        cmdmgr = self._cmdmgr
        if cmdmgr is None:
            lines.append('\tCommands have not been loaded.')
        else:
            for category in cmdmgr.categories:
                lines.append('\t{} COMMANDS'.format(category.upper()))

                # To deduplicate commands with the same name that provide
                # different interfaces (but should have the same docs), map
                # command names to commands.
                cmds = {}
                for cmd in cmdmgr.all_commands:
                    if category == cmd.category:
                        cmds[cmd.name] = cmd

                for cmdname,cmd in sorted(cmds.items()):
                    lines.append('\t\t{}  \t{}'.format(', '.join((cmd.name,)+cmd.aliases),
                                                       cmd.description))
                lines.append('')
        return finalize_lines(lines)

    @commands.setter
    def commands(self, cmdmgr):
        self._cmdmgr = cmdmgr

    def command(self, name):
        """Return help text for command"""
        cmd = self._cmdmgr[name]

        def takes_value(argspec):
            if argspec.get('action') in ('store_true', 'store_false', 'store_const'):
                return False  # Boolean option
            if 'nargs' not in argspec:
                return True
            nargs = argspec['nargs']
            return not isinstance(nargs, int) or nargs > 0

        def arg_dest(argspec):
            if 'metavar' in argspec:
                dest = argspec['metavar'].upper()
            elif 'dest' in argspec:
                dest = argspec['dest'].upper()
            elif argspec['names'][0][0] == '-':
                dest = argspec['names'][0].lstrip('-').upper()
            else:
                return None

            if 'nargs' in argspec and argspec['nargs'] in ('*', '?'):
                return '[<%s>]' % dest
            else:
                return '<%s>' % dest

        lines = [cmd.name.upper()]

        log.debug('Generating help text for %s', cmd.name)
        names = ', '.join((cmd.name,) + cmd.aliases)
        lines = [names + ' - \t' + cmd.description]
        lines.append('')

        if cmd.usage:
            lines.append('USAGE')
            lines += ['\t' + u for u in cmd.usage]
            lines.append('')

        if cmd.argspecs:
            lines.append('ARGUMENTS')
            lines_args = []
            for argspec in cmd.argspecs:
                argline = '\t' + ','.join(argspec['names'])
                if takes_value(argspec):
                    dest = arg_dest(argspec)
                    if dest is not None:
                        argline += ' %s' % dest

                argline += '  \t' + argspec['description']

                def stringify_default(default):
                    dflt = default() if callable(default) else default

                    if not isinstance(dflt, str) and isinstance(dflt, abc.Sequence):
                        return ' '.join(dflt)
                    else:
                        return str(dflt)

                if 'default_description' in argspec:
                    argline += ' (default: {})'.format(
                        stringify_default(argspec['default_description'])
                    )
                elif 'default' in argspec:
                    argline += ' (default: {})'.format(
                        stringify_default(argspec['default'])
                    )

                lines_args.append(argline)
            lines += lines_args
            lines.append('')

        if cmd.examples:
            lines.append('EXAMPLES')
            lines += ['\t' + e for e in cmd.examples]
            lines.append('')

        for name,text in sorted(cmd.more_sections.items()):
            lines.append(name.upper())
            if callable(text):
                text = text()
            lines += ['\t' + line for line in text]
            lines.append('')

        return finalize_lines(lines)

    @property
    def keymap(self):
        """Must be set to a KeyMap object; provides a help text"""

        # Make sure the TUI (and ergo the keymap) is loaded
        from .tui import main

        km = self._keymap
        lines = []

        if km is None:
            lines.append('Keybindings have not been loaded.')
        else:
            def stringify(s):
                return ' '.join(s) if not isinstance(s, str) else s

            for context in sorted(km.contexts, key=lambda c: '' if c is None else c):
                if context is None:
                    lines.append('GENERAL KEYBINDINGS')
                else:
                    lines.append('{} KEYBINDINGS'.format(context.upper()))

                keymap = ((key, stringify(action)) for key,action in km.map(context))

                # Sort by command
                from natsort import (natsort_keygen, natsorted, ns)
                get_cmd = natsort_keygen(key=lambda pair: pair[1], alg=ns.IGNORECASE)
                for key, action in natsorted(keymap, key=get_cmd):
                    lines.append('\t{}  \t{}'.format(key, action))
                lines.append('')
        return finalize_lines(lines)

    @keymap.setter
    def keymap(self, keymap):
        self._keymap = keymap

    @property
    def filter(self):
        """Provide help text for arguments to TorrentFilter"""
        lines = [
            'FILTERING TORRENTS, FILES AND PEERS',
            ('\tCommands that accept FILTER arguments for torrents, files or peers '
             'are applied to torrents, files or peers that match these arguments.'),
            '',
            '\tThere are two kinds of filters:',
            "\t\t- \tBoolean filters stand on their own (e.g. 'downloading')",
            "\t\t- \tComparative filters match against values (e.g. 'seeds>20').",
            '',
            '\tThe syntax of comparative filters is: [[<FILTER NAME>] <OPERATOR>] <VALUE>',
            '',
            ("\tBesides the standard operators (=, !=, >, <, >=, <=), '~' matches "
             "if the torrent's value contains VALUE."),
            "\tExample: 'name~foo' matches all torrents with 'foo' in their name.",
            '',
            ("\tIf FILTER NAME is omitted, it defaults to 'name'.  If OPERATOR is "
             "omitted, it defaults to '~'."),
            "\tExample: 'foo' is the same as '~foo' is the same as 'name~foo'.",
            '',
            ('\tSpaces before the filter name are ignored.  If there is a space '
             'between the filter name and the operator, all spaces at the '
             'start and end of VALUE are removed, otherwise they are preserved.  '),
            "\tExample: 'name = foo  ' matches 'foo'; 'name= foo  ' matches ' foo  '.",
            '',
            "\tAll filters can be inverted by prepending '!'.",
            ("\tExample: 'name!=foo' is the same as '!name=foo'; "
             "'!name!=foo' is the same as 'name=foo'."),
            '',
            ('\tWhen matching strings, matches are case-sensitive if VALUE includes '
             'upper-case characters, otherwise it is case-insensitive.'),
            '',
            ("\tWhen matching numbers, the unit prefixes 'k', 'M', 'G', 'T' and "
             "their binary counterparts 'Ki', 'Mi', 'Gi', 'Ti' are supported.  "
             "The case of unit prefixes is ignored."),
            ("\tExample: 'size>1mi' is the same as 'size>1048576' (1 Mebibyte); "
             "'size>1m' is the same as 'size>1000000' (1 Megabyte)"),
            '',
            ("\tFilters can be combined with the operators '&' (logical AND) "
             "and '|' (logical OR).  Multiple FILTER arguments are combined with '|'."),
            "\tExample: 'name=foo paused' is the same as 'name=foo|paused'.",
        ]

        from .client.filters.tfilter import SingleTorrentFilter
        from .client.filters.ffilter import SingleTorrentFileFilter
        from .client.filters.pfilter import SingleTorrentPeerFilter

        for caption,filt in (('TORRENT FILTERS', SingleTorrentFilter),
                             ('TORRENT FILE FILTERS', SingleTorrentFileFilter),
                             ('TORRENT PEER FILTERS', SingleTorrentPeerFilter)):
            lines += ['', '\t%s' % caption]
            lines.append('\t\tBOOLEAN FILTERS')
            for fname,f in sorted(filt.BOOLEAN_FILTERS.items()):
                lines.append('\t\t\t{} \t{}'.format(', '.join((fname,)+f.aliases), f.description))
            lines += ['', '\t\tCOMPARATIVE FILTERS']
            for fname,f in sorted(filt.COMPARATIVE_FILTERS.items()):
                lines.append('\t\t\t{} \t{}'.format(', '.join((fname,)+f.aliases), f.description))

        return finalize_lines(lines)

    @property
    def rcfile(self):
        """Provide help text for rc file"""
        lines = [
            'RC FILES',
            ("\tAn rc file contains a list of arbitrary commands.  Lines starting "
             "with '#' (or more precisely '\s*#') are ignored."),
            '',
            ("\tThe default rc file path is '$XDG_CONFIG_HOME/{APPNAME}/rc'.  "
             "XDG_CONFIG_HOME defaults to '~/.config'.  A different path can be "
             "provided with the '--rcfile' option.  An existing rc file at the "
             "default path can be ignored with the '--norcfile' option."),
            '',
            ('\tCommands in an rc file are called during startup before the '
             'commands given on the command line.'),
            '',
            ("\tTUI commands (e.g. 'tab' or 'bind') in an rc file are ignored "
             "in CLI mode."),
        ]
        return finalize_lines(lines)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from ..settings import is_server_setting
from ..commands import is_op as is_cmd_op


class UIGuessError(Exception):
    pass

def guess_ui(clicmds, cmdmgr):
    """Guess desired user interface based on CLI commands

    Return 'tui' or 'cli'
    """
    if not clicmds:
        return 'tui'

    log.debug('Guessing whether TUI or CLI is wanted:')
    tui_needed = False
    cli_needed = False
    guess = 'cli'
    for cmdline in cmdmgr.split_cmdchain(clicmds):
        if is_cmd_op(cmdline):
            continue

        cmdname = cmdline[0]
        debugmsg = '  %s: ' % cmdname

        tuicmd = cmdmgr.get_cmdcls(cmdname, interface='tui')
        clicmd = cmdmgr.get_cmdcls(cmdname, interface='cli')
        if tuicmd is None is clicmd:
            debugmsg += 'unknown command - not guessing'

        # Does command provide only one interface?
        elif tuicmd is None:
            debugmsg += 'no support for TUI - demanding CLI'
            cli_needed = True
        elif clicmd is None:
            debugmsg += 'no support for CLI - demanding TUI'
            tui_needed = True

        elif tuicmd.category == clicmd.category == 'torrent':
            # Torrent commands (start, stop, list, ...) inhibit the TUI
            debugmsg += 'torrent command - guessing CLI'
            guess = 'cli'

        # Some 'set' commands should enforce the tui or cli, other 'set'
        # commands shouldn't care.
        elif cmdline[0] == 'set' and len(cmdline) >= 2 or \
             cmdline[0] == 'reset' and len(cmdline) >= 2:
            setting = cmdline[1]
            if setting.startswith('tui.'):
                debugmsg += 'TUI setting: %r - guessing TUI' % setting
                guess = 'tui'
            elif is_server_setting(setting):
                debugmsg += 'server setting: %r - guessing CLI' % setting
                guess = 'cli'
            else:
                debugmsg += 'other setting: %r - guessing TUI' % setting
                guess = 'tui'

        elif clicmd is not None:
            debugmsg += 'CLI supported - guessing CLI'
            guess = 'cli'

        else:
            debugmsg += 'no guess'
        log.debug(debugmsg)

    if cli_needed and tui_needed:
        raise UIGuessError('Unable to guess user interface')
    elif tui_needed:
        ui = 'tui'
    elif cli_needed:
        ui = 'cli'
    else:
        ui = guess

    log.debug('Guessed wanted UI: %s', ui)
    return ui



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from functools import partial


class SortSpecBase():
    def __init__(self, *keyfuncs, description, aliases=()):
        self._keyfuncs = keyfuncs
        self.description = description
        self.aliases = aliases

    def __call__(self, items, reverse=False, inplace=False, item_getter=lambda item: item):
        if not items:
            return items

        for keyfunc in self._keyfuncs:
            def key_getter(item):
                return keyfunc(item_getter(item))

            if inplace:
                items[:] = sorted(items, key=key_getter, reverse=reverse)
            else:
                items = sorted(items, key=key_getter, reverse=reverse)

        return items


class SorterBase():
    SORTSPECS = NotImplemented
    DEFAULT_SORT = None

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)

        # Map aliases to their original name
        obj._aliases = {alias: sname
                       for sname,s in cls.SORTSPECS.items()
                       for alias in s.aliases}
        return obj

    def __init__(self, sortstrings=()):
        sortspecs = []
        sortfuncs = []
        strings = []         # String representations of sortspecs

        # Go through items in reverse because to want to deduplicate sort orders
        # while keeping the most recent one.
        for sortstring in reversed(sortstrings):
            if sortstring[0] in ('!', '.'):
                sortspecname, reverse = sortstring[1:], True
            else:
                sortspecname, reverse = sortstring, False

            # Resolve alias
            if sortspecname in self._aliases:
                sortspecname = self._aliases[sortspecname]

            if sortspecname not in self.SORTSPECS:
                raise ValueError('Unknown sort order: {!r}'.format(sortspecname))
            else:
                sortspec = self.SORTSPECS[sortspecname]
                if sortspec not in sortspecs:
                    sortfunc = partial(sortspec, reverse=reverse)
                    sortspecs.insert(0, sortspec)
                    sortfuncs.insert(0, sortfunc)
                    strings.insert(0, ('!' if reverse else '') + sortspecname)
        self._strings = tuple(strings)

        # Unless we already sort by DEFAULT_SORT, insert it as the first one.
        if self.DEFAULT_SORT is not None:
            default_sortspec = self.SORTSPECS[self.DEFAULT_SORT]
            if default_sortspec not in sortspecs:
                sortfuncs.insert(0, default_sortspec)
                sortspecs.insert(0, default_sortspec)

        self._sortspecs = sortspecs
        self._sortfuncs = sortfuncs

    def apply(self, items, inplace=False, item_getter=lambda item: item):
        """Sort sequence `items`

        item_getter: Callable that gets an item of `items` and returns an
                     object that can be sorted with any of the sorters
                     specified in the SORTSPECS variable. (This allows for
                     sorting of widgets as long as they can provide a sortable
                     object.)
        inplace: Modify `items` if True, otherwise return a new list
        """
        import time
        start_time = time.monotonic()

        for sorter in self._sortfuncs:
            items = sorter(items, inplace=inplace, item_getter=item_getter)

        log.debug('-> Sorted %d items by %s in %.3fms',
                  len(items), self, (time.monotonic()-start_time)*1e3)

        if not inplace:
            return items

    def __add__(self, other):
        cls = type(self)
        if not isinstance(other, cls):
            return NotImplemented
        else:
            return cls(self._strings + other._strings)

    def __str__(self):
        return ','.join(self._strings)

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, str(self))
